import libsbml
import warnings
from sympy.logic.boolalg import And, Or, Not
from sympy.core.relational import StrictGreaterThan, StrictLessThan, GreaterThan, LessThan, Eq
from sympy import sympify, symbols
import os

DIR = os.path.dirname(os.path.realpath(__file__))
PATH = os.path.join(DIR, '../../../examples/models/regulation/e_coli_core.xml')

ASTNODE_BOOLEAN_OPERATORS = {
    libsbml.AST_LOGICAL_AND: And,
    libsbml.AST_LOGICAL_OR: Or,
    libsbml.AST_LOGICAL_NOT: Not}

ASTNODE_RELATIONAL_OPERATORS = {
    libsbml.AST_RELATIONAL_EQ: Eq,
    libsbml.AST_RELATIONAL_LEQ: LessThan,
    libsbml.AST_RELATIONAL_GEQ: GreaterThan,
    libsbml.AST_RELATIONAL_LT: StrictLessThan,
    libsbml.AST_RELATIONAL_GT: StrictGreaterThan
}

ASTNODE_VALUES = {
    libsbml.AST_INTEGER: lambda x: int(x.getValue()),
    libsbml.AST_REAL: lambda x: x.getValue(),
    libsbml.AST_REAL_E: lambda x: x.getValue(),
    libsbml.AST_RATIONAL: lambda x: x.getValue(),
    libsbml.AST_CONSTANT_FALSE: 0,
    libsbml.AST_CONSTANT_TRUE: 1
}

ASTNODE_NAME = libsbml.AST_NAME


def readMathNode(ast_node):

    """
    Reads and parses a node of type math ASTNode into a boolean sympy expression.

    :param ast_node: ASTNode, math ASTNode from ft.getMath()
    :return: sympy expression object or None
    """

    def _process_rule(node):

        node_type = node.getType()

        if node_type in ASTNODE_BOOLEAN_OPERATORS:

            # TODO: If op == Not and node.getNumChildren() > 1, something is wrong in the sbml file

            op = ASTNODE_BOOLEAN_OPERATORS[node_type]
            args = [_process_rule(node.getChild(child)) for child in range(node.getNumChildren())]

            # noinspection PyArgumentList
            sp = op(*args, evaluate=False)

            return sp

        elif node_type in ASTNODE_RELATIONAL_OPERATORS:

            # TODO: If node.getNumChildren() > 2, something is wrong in the sbml file
            # TODO: Process equalities

            op = ASTNODE_RELATIONAL_OPERATORS[node_type]
            args = [_process_rule(node.getChild(child)) for child in range(node.getNumChildren())]

            # noinspection PyArgumentList
            sp = op(*args, evaluate=False)

            return sp

        elif node_type == ASTNODE_NAME:

            # TODO: Verify if the name appears in the list of inputs and species
            # TODO: see if it is possible to change the name of the node. If not, create a aliases_map to find the
            #  variable

            return symbols(node.getName())

        elif node_type in ASTNODE_VALUES:

            # TODO: It should be a state for the previous variable, zero (which means not variable), or one (which
            #  means that the variable directly influences the outcome of the outputs)

            return sympify(ASTNODE_VALUES[node_type](node))

        else:

            return

    try:
        return _process_rule(ast_node)
    except Exception as e:
        print(e)
        print('Moving on')
        return

doc = libsbml.readSBML(PATH)

model = doc.getModel()

model_qual = model.getPlugin("qual")

model_id = model.getIdAttribute()

valid_id = libsbml.SyntaxChecker.isValidSBMLSId(model_id)

model_name = model.getName()

level = model.getLevel()

version = model.getVersion()

has_history = model.isSetModelHistory()

if has_history:

    history = model.getModelHistory()

    if history.isSetCreatedDate():
        created = history.getCreatedDate()

    creators = history.getListCreators()

# TODO: parsing compartments to create metabolite aliases with the respective compartment annotation. Either way,
#  the ID reflect the compartment and be equal to the one in the CBM model
compartments = {}
for compartment in model.getListOfCompartments():
    compartments[compartment.getIdAttribute()] = compartment.getName()

# TODO: parsing qual species IDs
# TODO: parsing states in notes

qual_species = {}
for qs in model_qual.getListOfQualitativeSpecies():
    qs_id = qs.getIdAttribute()
    name = qs.getName()
    if not name:
        name = qs_id
    compartment = qs.getCompartment()
    constant = qs.getConstant()
    initialLevel = qs.getInitialLevel()
    maxlevel = qs.getMaxLevel()
    notes = qs.getNotesString()
    qual_species[qs.getIdAttribute()] = (qs_id, name, compartment, constant, initialLevel, maxlevel, notes)

interactions = {}

for qt in model_qual.getListOfTransitions():

    qt_id = qt.getIdAttribute()
    name = qt.getName()
    if not name:
        name = qt_id

    # regulators
    inputs = {}
    for reg in qt.getListOfInputs():
        reg_id = reg.getIdAttribute()
        reg_name = reg.getName()
        specie_id = qual_species[reg.getQualitativeSpecies()][0]
        inputs[specie_id] = (reg_id, reg_name, specie_id)

    # TODO: if a given variable is the output of more than one transition, these transitions must be combined into a
    #  single one. Alternatively, launch a warning to repair the sbml as we do not handle this

    # TODO: Caution, there may be more than one output for a given transition

    # targets
    outputs = {}
    for tg in qt.getListOfOutputs():
        tg_id = tg.getIdAttribute()
        tg_name = tg.getName()
        specie_id = qual_species[tg.getQualitativeSpecies()][0]
        outputs[specie_id] = (tg_id, tg_name, specie_id)

    function_terms = {'default': ('', qt.getDefaultTerm().getResultLevel())}

    for ft in qt.getListOfFunctionTerms():

        ft_result = ft.getResultLevel()

        ft_rule = readMathNode(ft.getMath())

        if ft_rule is None:
            ft_result = ''

        function_terms['interaction_' + str(ft_result)] = (ft_rule, ft_result)

    interactions[qt_id] = (qt_id, name, inputs, outputs, function_terms)

# TODO: workflow:
#  Create empty model ->
#  Add regulatory variables (parsing) ->
#  Add regulatory interactions from sympy object


""" JMetal Problems 
"""
from jmetal.core.solution import Solution, IntegerSolution
from jmetal.core.problem import Problem
from mewpy.optimization.ea import SolutionInterface, dominance_test
from typing import Tuple, List
import random
import warnings

# define EA representation for OU
IntTupple = Tuple[int]


class KOSolution(Solution[int], SolutionInterface):
    """ Class representing a KO solution """

    def __init__(self, lower_bound: int, upper_bound: int, number_of_variables: int, number_of_objectives: int):
        super(KOSolution, self).__init__(number_of_variables,
                                         number_of_objectives)
        self.lower_bound = lower_bound
        self.upper_bound = upper_bound

    def __eq__(self, solution) -> bool:
        if isinstance(solution, self.__class__):
            return self.variables.sort() == solution.variables.sort()
        return False

    # JMetal consideres all problems as minimization
    # Based on pareto dominance

    def __gt__(self, solution) -> bool:
        if isinstance(solution, self.__class__):
            return dominance_test(self, solution, maximize=False) == 1
        return False

    def __lt__(self, solution) -> bool:
        if isinstance(solution, self.__class__):
            return dominance_test(self, solution, maximize=False) == -1
        return False

    def __ge__(self, solution) -> bool:
        if isinstance(solution, self.__class__):
            return dominance_test(self, solution, maximize=False) != -1
        return False

    def __le__(self, solution) -> bool:
        if isinstance(solution, self.__class__):
            return dominance_test(self, solution, maximize=False) != 1
        return False

    def __copy__(self):
        new_solution = KOSolution(
            self.lower_bound,
            self.upper_bound,
            self.number_of_variables,
            self.number_of_objectives)
        new_solution.objectives = self.objectives[:]
        new_solution.variables = self.variables[:]
        new_solution.constraints = self.constraints[:]
        new_solution.attributes = self.attributes.copy()

        return new_solution

    def get_representation(self):
        """
        Returns a set representation of the candidate 
        """
        return set(self.variables)

    def get_fitness(self):
        """
        Returns the candidate fitness list
        """
        return self.objectives

    def __str__(self):
        return " ".join((self.variables))

    
    def __getitem__(self,idx):
        return self.objectives[idx]

    def __len__(self):
        return len(self.objectives)


class OUSolution(Solution[IntTupple], SolutionInterface):
    """
    Class representing a Over/Under expression solution.
    """

    def __init__(self, lower_bound: List[int], upper_bound: List[int], number_of_variables: int, number_of_objectives: int):
        super(OUSolution, self).__init__(number_of_variables,
                                         number_of_objectives)
        self.upper_bound = upper_bound
        self.lower_bound = lower_bound

    def __eq__(self, solution) -> bool:
        if isinstance(solution, self.__class__):
            return self.variables.sort() == solution.variables.sort()
        return False

    # JMetal consideres all problems as minimization

    def __gt__(self, solution) -> bool:
        if isinstance(solution, self.__class__):
            return dominance_test(self, solution, maximize=False) == 1
        return False

    def __lt__(self, solution) -> bool:
        if isinstance(solution, self.__class__):
            return dominance_test(self, solution, maximize=False) == -1
        return False

    def __ge__(self, solution) -> bool:
        if isinstance(solution, self.__class__):
            return dominance_test(self, solution, maximize=False) != -1
        return False

    def __le__(self, solution) -> bool:
        if isinstance(solution, self.__class__):
            return dominance_test(self, solution, maximize=False) != 1
        return False

    def __copy__(self):
        new_solution = OUSolution(
            self.lower_bound,
            self.upper_bound,
            self.number_of_variables,
            self.number_of_objectives
        )
        new_solution.objectives = self.objectives[:]
        new_solution.variables = self.variables[:]
        new_solution.constraints = self.constraints[:]
        new_solution.attributes = self.attributes.copy()

        return new_solution

    def get_fitness(self):
        """
        Returns the candidate fitness list
        """
        return self.objectives

    def __str__(self):
        return " ".join((self.variables))


    def __getitem__(self,idx):
        return self.objectives[idx]

    def __len__(self):
        return len(self.objectives)



class JMetalKOProblem(Problem[KOSolution]):

    def __init__(self, problem):
        self.problem = problem
        self.number_of_objectives = len(self.problem.fevaluation)
        self.obj_directions = []
        self.obj_labels = []
        for f in self.problem.fevaluation:
            self.obj_labels.append(str(f))
            if f.maximize:
                self.obj_directions.append(self.MAXIMIZE)
            else:
                self.obj_directions.append(self.MINIMIZE)

    def create_solution(self) -> KOSolution:
        solution = self.problem.generator(random, None)
        new_solution = KOSolution(
            self.problem.bounder.lower_bound,
            self.problem.bounder.upper_bound,
            len(solution),
            self.problem.number_of_objectives)
        new_solution.variables = list(solution)
        return new_solution

    def get_constraints(self, solution):
        return self.problem.decode(set(solution.variables))

    def evaluate(self, solution: KOSolution) -> KOSolution:
        candidate = set(solution.variables)
        p = self.problem.evaluate_solution(candidate)
        for i in range(len(p)):
            # JMetalPy only deals with minimization problems
            if self.obj_directions[i] == self.MAXIMIZE:
                solution.objectives[i] = -1 * p[i]
            else:
                solution.objectives[i] = p[i]
        return solution

    def get_name(self) -> str:
        return self.problem.get_name()


class JMetalOUProblem(Problem[OUSolution]):

    def __init__(self, problem):
        self.problem = problem
        self.number_of_objectives = len(self.problem.fevaluation)
        self.obj_directions = []
        self.obj_labels = []
        for f in self.problem.fevaluation:
            self.obj_labels.append(str(f))
            if f.maximize:
                self.obj_directions.append(self.MAXIMIZE)
            else:
                self.obj_directions.append(self.MINIMIZE)

    def create_solution(self) -> OUSolution:
        # uses the super class generator
        solution = self.problem.generator(random, None)

        new_solution = OUSolution(
            self.problem.bounder.lower_bound,
            self.problem.bounder.upper_bound,
            len(solution),
            self.problem.number_of_objectives)
        new_solution.variables = list(solution)
        return new_solution

    def get_constraints(self, solution):
        return self.problem.decode(set(solution.variables))

    def evaluate(self, solution: KOSolution) -> KOSolution:
        candidate = set(solution.variables)
        p = self.problem.evaluate_solution(candidate)
        for i in range(len(p)):
            # JMetalPy only deals with minimization problems
            if self.obj_directions[i] == self.MAXIMIZE:
                solution.objectives[i] = -1 * p[i]
            else:
                solution.objectives[i] = p[i]
        return solution

    def get_name(self) -> str:
        return self.problem.get_name()



from .problem import Strategy, AbstractKOProblem, AbstractOUProblem
from .reactions import RKOProblem, ROUProblem
from .genes import GKOProblem, GOUProblem
from .gecko import GeckoKOProblem, GeckoOUProblem



from warnings import warn
from mewpy.utils.parsing import generic_regulatory_rule_parser
from mewpy.regulation.regulatory_variable import RegulatoryVariable


class RegulatoryInteraction(object):

    def __init__(self, identifier,
                 rule,
                 target,
                 regulators=None,
                 model=None,
                 aliases=None):

        if not aliases:
            aliases = []

        self._id = None
        self._aliases = None
        self.old_rule = None
        self._rule = None
        self._elements = None
        self._target = None
        self._regulators = None
        self._model = None

        self.__build_regulatory_interaction__(identifier,
                                              aliases,
                                              rule,
                                              target,
                                              regulators,
                                              model)

    def __str__(self):
        return self.target.id + ' ' + self.rule

    def __repr__(self):

        return self.id

    # def __repr__(self):
    #
    #     return '{}({}, {}, {}) at {}'.format(self.__class__.__name__, self._id, self._regulatory_interaction_rule,
    #     self._target, id(self))

    def __delete__(self, instance):
        del self.id
        del self.rule
        del self.elements
        del self.old_rule
        del self.target
        del self.regulators
        del self.model

    def __build_regulatory_interaction__(self,
                                         identifier,
                                         aliases,
                                         rule,
                                         target,
                                         regulators,
                                         model):

        if not identifier:
            raise BaseException("An ID must be provided")

        if not isinstance(identifier, str):
            raise TypeError("The ID must be a string. Use str() to convert")

        if not rule:
            def reg_rule_msg(message):
                warn(message, UserWarning)

            reg_rule_msg(
                "The regulatory interaction rule is empty for the target {}".format(target.id))

            rule = ''

        if regulators:

            if not isinstance(regulators, dict) or \
                    not isinstance(regulators[list(regulators.keys())[0]], RegulatoryVariable):
                raise TypeError(
                    "Regulators must be a dict of RegulatoryVariables objects")

            self._regulators = regulators

        else:
            self._regulators = {}

        self.id = identifier
        self.aliases = aliases

        self.target = target

        self.model = model

        self.rule = rule

    @classmethod
    def from_str(cls, identifier,
                 rule,
                 target,
                 target_aliases=None,
                 regulators=None,
                 model=None):

        if not isinstance(identifier, str):
            raise TypeError("IDs must be a list of strings")

        if not isinstance(rule, str):
            raise TypeError(
                "Rules must be a dict object of list of strings. Use list() to convert")

        if isinstance(target, str):

            if model is not None:

                tar_var = model.__add_target__(identifier=target,
                                               name=target,
                                               aliases=target_aliases,
                                               regulators=regulators)

            else:

                tar_var = RegulatoryVariable(identifier=target,
                                             name=target,
                                             aliases=target_aliases,
                                             is_target=True)

            return cls(identifier=identifier,
                       rule=rule,
                       target=tar_var,
                       model=model)

        elif isinstance(target, RegulatoryVariable):

            return cls(identifier=identifier,
                       rule=rule,
                       target=target,
                       model=model)

        else:
            raise TypeError(
                "Target must be a string or a RegulatoryVariable object")

    @classmethod
    def from_dict(cls, identifiers,
                  rules,
                  targets_aliases=None,
                  model=None):

        if not isinstance(identifiers, list):
            raise TypeError("IDs must be a list of strings")

        if not isinstance(identifiers[0], str):
            raise TypeError("IDs must be a list of strings")

        if not isinstance(rules, dict):
            raise TypeError(
                "Rules must be a dict object of list of strings. Use dict() to convert")

        if not isinstance(rules[list(rules.keys())[0]], str):
            raise TypeError(
                "Rules must be a dict object of list of strings. Use list() to convert")

        if not targets_aliases:

            if not isinstance(targets_aliases, dict):
                raise TypeError(
                    "The target_aliases must be a dict object. Use dict() to convert")

            if not isinstance(targets_aliases[list(targets_aliases.keys())[0]], list):
                raise TypeError(
                    "The target_aliases must be a dict object of lists of strings. Use list() to convert")

            if not isinstance(targets_aliases[list(targets_aliases.keys())[0]][0], str):
                raise TypeError(
                    "The target_aliases must be a dict object of lists of strings. Use str() to convert")

        else:
            targets_aliases = {}

        reg_inters = []

        for i, (key, val) in enumerate(rules.items()):

            if key in targets_aliases:
                aliases = targets_aliases[key]
            else:
                aliases = []

            reg_inter = cls.from_str(identifiers[i],
                                     rule=val,
                                     target=key,
                                     target_aliases=aliases,
                                     regulators=None,
                                     model=model)

            reg_inters.append(reg_inter)

        return reg_inters

    @property
    def id(self):
        return getattr(self, '_id', None)

    @property
    def aliases(self):
        return getattr(self, '_aliases', None)

    @property
    def rule(self):
        return getattr(self, '_rule', None)

    @property
    def elements(self):
        return getattr(self, '_elements', None)

    @property
    def target(self):
        return getattr(self, '_target', None)

    @property
    def regulators(self):
        return getattr(self, '_regulators', None)

    @property
    def model(self):
        return getattr(self, '_model', None)

    @id.setter
    def id(self, value):

        if value == self._id:
            pass

        elif not isinstance(value, str):
            raise TypeError("The ID must be a string. Use str() to convert")

        else:

            if self.model:

                if value in self.model.regulatory_interactions:
                    def reg_id(message):
                        warn(message, UserWarning, stacklevel=2)

                    reg_id("The ID already exists in the dictionary of regulatory interactions of the regulatory "
                           "model. Ignoring the alteration")

                else:

                    if self._id in self.model.regulatory_interactions:
                        del self.model.regulatory_interactions[self._id]
                        self.model.regulatory_interactions[value] = self

                    if self.regulators:
                        for reg in self.regulators_gen():
                            if self._id in reg.regulatory_interactions:
                                del reg.regulatory_interactions[self._id]
                                reg.regulatory_interactions[value] = self

                    if self.target:
                        if self._id in self.target.regulatory_interactions:
                            del self.target.regulatory_interactions[self._id]
                            self.target.regulatory_interactions.regulatory_interactions[value] = self

                    self._id = value

            else:

                if self.regulators:
                    for reg in self.regulators_gen():
                        if self._id in reg.regulatory_interactions:
                            del reg.regulatory_interactions[self._id]
                            reg.regulatory_interactions[value] = self

                if self.target:
                    if self._id in self.target.regulatory_interactions:
                        del self.target.regulatory_interactions[self._id]
                        self.target.regulatory_interactions.regulatory_interactions[value] = self

                self._id = value

    @aliases.setter
    def aliases(self, value):

        if value == self._aliases:
            pass

        elif not value:
            pass

        else:

            if not isinstance(value, list):
                raise TypeError("The aliases must be a list. Use list() to convert")

            if value and not isinstance(value[0], str):
                raise TypeError("The aliases must be a list. Use list() to convert")

            if not self._aliases:
                self._aliases = []

            for val in value:
                if val not in self._aliases:
                    self._aliases = self._aliases + [val]

    @rule.setter
    def rule(self, value):

        if value == self._rule:
            pass

        elif not isinstance(value, str):
            raise TypeError(
                "Regulatory interaction rule must be a string. Use str() to convert")

        else:

            parser = generic_regulatory_rule_parser(value)

            # return parsed_str (str), 1
            # exp_elements (list), 2

            self.old_rule = value

            self._rule = parser[0]
            self._elements = parser[1]

            if self.regulators:

                for key in self.regulators:
                    self.regulators[key].regulatory_interactions[self.id] = self
                    self.regulators[key].is_regulator = True
                    self.regulators[key].model = self.model
                    self.regulators[key].targets[self.target.id] = self.target
                    self.target.regulators[key] = self.regulators[key]

    @elements.setter
    def elements(self, value):

        raise BaseException("regulatory_interaction_elements property cannot be changed. "
                            "please create/set a new regulatory interaction rule.")

    @target.setter
    def target(self, value):

        if value == self._target:
            pass

        elif not value:
            raise BaseException("A target must provided as a RegulatoryVariable object")

        elif not isinstance(value, RegulatoryVariable):
            raise TypeError("Target must be a RegulatoryVariable object")

        else:

            self._target = value

            self._target.is_target = True
            self._target.regulatory_interactions[self.id] = self
            self._target.model = self.model

    @regulators.setter
    def regulators(self, value):

        raise BaseException("Regulators property cannot be changed. "
                            "To change the regulators of this regulatory interacion, "
                            "please change the regulatory interaction rule.")

    @model.setter
    def model(self, value):

        # TODO: The setter should also remove the regulatory variable from the old model if exists

        from mewpy.regulation.regulatory_model import RegulatoryModel

        if value == self._model:
            pass

        elif not value:
            self._model = None

        elif not isinstance(value, RegulatoryModel):
            raise TypeError("Regulatory model must be a Regmodel object or None")

        else:

            # TODO: This should point all variables and interactions related with this one to the same model, but
            #  this might be a recursion problem

            if self.id not in value.regulatory_interactions:
                value.regulatory_interactions[self.id] = self
                self._model = value

                if self.target:
                    if self.target.id not in value.targets:
                        value.targets[self.target.id] = self.target
                    if self.target.id in value.regulators:
                        self.target.is_regulator = True

                if self.regulators:
                    for reg in self.regulators_gen():
                        if reg.id not in value.regulators:
                            value.regulators[reg.id] = reg
                        if reg.id in value.targets:
                            reg.is_target = True

            else:
                def model_id(message):
                    warn(message, UserWarning, stacklevel=2)

                model_id("The ID already exists in the dictionary of regulatory interactions of the regulatory "
                         "model. Ignoring the alteration")

    @id.deleter
    def id(self):
        del self._id

    @rule.deleter
    def rule(self):
        del self._rule

    @elements.deleter
    def elements(self):
        del self._elements

    @target.deleter
    def target(self):
        del self._target

    @regulators.deleter
    def regulators(self):
        del self._regulators

    @model.deleter
    def model(self):
        del self._model

    def get_target(self):

        return self.target

    def get_regulator(self, identifier):

        return self.regulators[identifier]

    def regulators_gen(self):

        return (value for value in self._regulators.values())

    def remove_from_model(self):

        if self.model is not None:

            self.model.remove_regulatory_interaction(self.id)

        else:
            def remove_from_model_reg(message):
                warn(message, UserWarning, stacklevel=2)

            remove_from_model_reg("The regulatory interaction is not bounded to any regulatory model")



import numpy as np
from sympy import postorder_traversal
from sympy.logic.boolalg import And, Or, Not, BooleanFalse, BooleanTrue
from sympy.core.relational import StrictGreaterThan, StrictLessThan, GreaterThan, LessThan
from sympy.core.numbers import Zero, One
from cobamp.core.linear_systems import GenericLinearSystem, VAR_CONTINUOUS, VAR_BINARY
from cobamp.core.optimization import LinearSystemOptimizer
from mewpy.simulation.simulation import SimulationResult
from mewpy.regulation import IntegratedModel

_SRFBA_TOL = 1E-10


class SRFBAModel(IntegratedModel):
    """

    SRFBAModel class object inherits everything from the IntegratedModel class.
    This object is the standard for simulating a SRFBA model by solving a boolean regulatory model together with
    the metabolic model as Mixed Integer Linear Problem (MILP) asynchronously
    It provides the simulate method where one can know all about the metabolic and regulatory analysis of SRFBA

    """

    def __init__(self,
                 identifier,
                 name=None,
                 cbm_model=None,
                 cbm_simulation_interface=None,
                 targets=None,
                 regulators=None,
                 regulatory_interactions=None,
                 initial_state=None):

        self._A = None
        self._lbs = None
        self._ubs = None
        self._a_lb = None
        self._a_ub = None
        self._variables = None
        self._variables_types = {'boolean': [],
                                 'continuous': []}
        self._problem = None

        # integrated model
        super().__init__(identifier,
                         name,
                         cbm_model,
                         cbm_simulation_interface,
                         targets,
                         regulators,
                         regulatory_interactions,
                         initial_state)

    @property
    def A(self):
        return getattr(self, '_A', None)

    @property
    def lbs(self):
        return getattr(self, '_lbs', None)

    @property
    def ubs(self):
        return getattr(self, '_ubs', None)

    @property
    def a_lb(self):
        return getattr(self, '_a_lb', None)

    @property
    def a_ub(self):
        return getattr(self, '_a_ub', None)

    @property
    def variables(self):
        return getattr(self, '_variables', None)

    @property
    def problem(self):
        return getattr(self, '_problem', None)

    def __add_column__(self, variable, lb=0, ub=1, v_type='boolean'):

        # adding column/variable to the problem (return it if already exists)
        # It returns the column index

        if variable in self.variables:
            return self.variables[variable]

        for row in self.A:
            row.append(0)
        self.lbs.append(lb)
        self.ubs.append(ub)
        self.variables[variable] = len(self.A[0]) - 1
        self._variables_types[v_type].append(len(self.A[0]) - 1)

        return len(self.A[0]) - 1

    def __add_row__(self, elements, a_lb=0, a_ub=1):

        # adding row to the problem and return its index in the A matrix

        new_row = [0] * len(self.A[0])
        for i, val in elements:
            new_row[i] = val
        self.A.append(new_row)
        self.a_lb.append(a_lb)
        self.a_ub.append(a_ub)

        return len(self.A) - 1

    def __decode_reaction_bounds__(self, reaction, reaction_bool_variable):

        # Relation between reaction boolean value and the reaction constrains
        # For that, the following reactions must be added
        # V - Y*Vmax < 0
        # V - Y*Vmin > 0
        # where V is the reaction in S matrix
        # where Y is the reaction boolean variable

        _lb, _ub = self.lbs[reaction], self.ubs[reaction]

        self.__add_row__([(reaction, 1), (reaction_bool_variable, -_ub)], -999999, 0)

        self.__add_row__([(reaction, 1), (reaction_bool_variable, -_lb)], 0, 999999)

    def __boolean_rule_decode__(self, variable, expression):

        # Results of sympy's expression parsing:
        # gene = X and 1 => X
        # gene = X or 1 => 1 (sympy's BooleanTrue)
        # gene = X and 0 => 0 (sympy's BooleanFalse)
        # gene = X or 0 => X
        # gene = 1 => 1 (sympy's one)
        # gene = 0 => 0 (sympy's zero)
        # gene = X > 10 => (X, 1 (sympy's number)), namely two args

        # if expression is atom and defines a variable always On or Off, add a On/Off row
        if expression.is_Atom:

            if expression.is_Boolean or expression.is_Number:

                if isinstance(expression, BooleanFalse) or isinstance(expression, Zero):

                    # add row and set mip bounds to 0;0
                    self.__add_row__([(variable, 1)], a_ub=0)
                    return

                elif isinstance(expression, BooleanTrue) or isinstance(expression, One):

                    # add row and set mip bounds to 1;1
                    self.__add_row__([(variable, 1)], a_lb=1)
                    return

                else:
                    print("Either the target result is undetermined or "
                          "something is wrong with expression: {}".format(str(expression)))

                    # add row and set mip bounds to 0;1
                    self.__add_row__([(variable, 1)])
                    return

            # elif expression.is_Number:
            #
            #     if isinstance(expression, Zero):
            #
            #         # add row and set mip bounds to 0;0
            #         self.__add_row__([(variable, 1)], a_ub=0)
            #         return
            #
            #     elif isinstance(expression, One):
            #
            #         # add row and set mip bounds to 1;1
            #         self.__add_row__([(variable, 1)], a_lb=1)
            #         return
            #
            #     else:
            #         print("Either the target result is undetermined or "
            #               "something is wrong with expression: {}".format(str(expression)))
            #
            #         # add row and set mip bounds to 0;1
            #         self.__add_row__([(variable, 1)])
            #         return

            else:

                # Else it must be a symbol or nothing
                # The rest of the function can handle it
                pass

        traversal_results = {}
        last_index = 0

        # postorder_traversal recursively iterates an expression in the reverse order
        # e.g. expr = A and (B or C)
        # list(postorder_traversal(expr)) # [C, B, A, B | C, A & (B | C)]
        for arg in postorder_traversal(expression):

            if isinstance(arg, Or):

                # Following Boolean algebra, an Or (a = b or c) can be translated as: a = b + c - b*c
                # Alternatively, an Or can be written as lb < b + c - a < ub

                # So, for the mid term expression a = b or c
                # We have therefore the equation: -2 <= 2*b + 2*c – 4*a <= 1

                # add Or variable / column
                # or_col_idx = self.__add_column__('mid_term_' + str(self.A.shape[1]))
                or_col_idx = self.__add_column__('mid_term_' + str(len(self.A[0])))

                # Or left and right operators
                op_l = traversal_results[arg.args[0]]
                op_r = traversal_results[arg.args[1]]

                # add Or row and set mip bounds to -2;1
                self.__add_row__([(or_col_idx, -4), (op_l, 2), (op_r, 2)], a_lb=-2, a_ub=1)
                traversal_results[arg] = or_col_idx
                last_index = or_col_idx

                # building a nested Or subexpression
                for sub_arg in range(2, len(arg.args)):
                    # add Or variable / column
                    # or_col_idx = self.__add_column__('mid_term_' + str(self.A.shape[1]))
                    or_col_idx = self.__add_column__('mid_term_' + str(len(self.A[0])))

                    # Or left (last Or) and right operators
                    op_l = traversal_results[arg]
                    op_r = traversal_results[arg.args[sub_arg]]

                    # add Or row and set mip bounds to -2;1
                    self.__add_row__([(or_col_idx, -4), (op_l, 2), (op_r, 2)], a_lb=-2, a_ub=1)
                    traversal_results[arg] = or_col_idx
                    last_index = or_col_idx

            elif isinstance(arg, And):

                # Following Boolean algebra, an And (a = b and c) can be translated as: a = b*c
                # Alternatively, an And can be written as lb < b + c - a < ub

                # So, for the mid term expression a = b and c
                # We have therefore the equation: -1 <= 2*b + 2*c – 4*a <= 3

                # add And variable / column
                # and_col_idx = self.__add_column__('mid_term_' + str(self.A.shape[1]))
                and_col_idx = self.__add_column__('mid_term_' + str(len(self.A[0])))

                # And left and right operators
                op_l = traversal_results[arg.args[0]]
                op_r = traversal_results[arg.args[1]]

                # add And row and set mip bounds to -1;3
                self.__add_row__([(and_col_idx, -4), (op_l, 2), (op_r, 2)], a_lb=-1, a_ub=3)
                traversal_results[arg] = and_col_idx
                last_index = and_col_idx

                # building a nested And subexpression
                for sub_arg in range(2, len(arg.args)):
                    # add And variable / column
                    # and_col_idx = self.__add_column__('mid_term_' + str(self.A.shape[1]))
                    and_col_idx = self.__add_column__('mid_term_' + str(len(self.A[0])))

                    # And left (last And) and right operators
                    op_l = traversal_results[arg]
                    op_r = traversal_results[arg.args[sub_arg]]

                    # add Or row and set mip bounds to -2;1
                    self.__add_row__([(and_col_idx, -4), (op_l, 2), (op_r, 2)], a_lb=-1, a_ub=3)
                    traversal_results[arg] = and_col_idx
                    last_index = and_col_idx

            elif isinstance(arg, Not):

                # Following Boolean algebra, an Not (a = not b) can be translated as: a = 1 - b
                # Alternatively, an Not can be written as a + b = 1

                # So, for the mid term expression a = not b
                # We have therefore the equation: 1 < a + b < 1

                # add Not variable / column
                # not_col_idx = self.__add_column__('mid_term_' + str(self.A.shape[1]))
                not_col_idx = self.__add_column__('mid_term_' + str(len(self.A[0])))

                # Not right operators
                op_r = traversal_results[arg.args[0]]

                # add Not row and set mip bounds to 1;1
                self.__add_row__([(not_col_idx, 1), (op_r, 1)], a_lb=1)
                traversal_results[arg] = not_col_idx
                last_index = not_col_idx

            elif isinstance(arg, StrictGreaterThan) or isinstance(arg, GreaterThan):

                # Following Propositional logic, a predicate (a => r > value) can be translated as: r - value > 0
                # Alternatively, flux predicate a => r>value can be a(value + tolerance - r_UB) + r < value + tolerance
                # & a(r_LB - value - tolerance) + r > r_LB

                # So, for the mid term expression a => r > value
                # We have therefore the equations: a(value + tolerance - r_UB) + r < value + tolerance
                # & a(r_LB - value - tolerance) + r > r_LB

                # In matlab: c * (Vmax - const) + v <= Vmax
                # In matlab: const <= c * (const - Vmin) + v

                # In matlab: compareVal = -compareVal - epsilon
                # In matlab: matrix_values = [vmax - compare_vale, 1] (-inf, vmax)
                # In matlab: matrix_values = [compare_vale - vmin, 1] (compareVal, inf)

                # add Greater variable / column
                # greater_col_idx = self.__add_column__('mid_term_' + str(self.A.shape[1]))
                greater_col_idx = self.__add_column__('mid_term_' + str(len(self.A[0])))

                # Greater left operators
                # op_l can be growth, ph, ...
                # otherwise, it can be a reaction or a boundary reaction associated with a metabolite
                # Either way, they should be already in the matrix

                if arg.args[0].is_Symbol:
                    op_l = traversal_results[arg.args[0]]
                    c_val = arg.args[1]

                else:
                    op_l = traversal_results[arg.args[1]]
                    c_val = arg.args[0]

                _lb, _ub = self.lbs[op_l], self.ubs[op_l]

                # add Greater row (a(value + tolerance - r_UB) + r < value + tolerance) and set mip bounds to
                # -inf;comparison_val
                self.__add_row__([(greater_col_idx, c_val + _SRFBA_TOL - _ub), (op_l, 1)], a_lb=-999999999,
                                 a_ub=c_val + _SRFBA_TOL)
                # self.__add_row__([(greater_col_idx, _ub + c_val), (op_l, 1)], a_lb=-999999, a_ub=_ub)

                # add Greater row (a(r_LB - value - tolerance) + r > r_LB) and set mip bounds to lb;inf
                self.__add_row__([(greater_col_idx, _lb - c_val - _SRFBA_TOL), (op_l, 1)], a_lb=_lb, a_ub=999999999)
                # self.__add_row__([(greater_col_idx, - c_val - _lb), (op_l, 1)], a_lb=-c_val, a_ub=999999)

                traversal_results[arg] = greater_col_idx
                last_index = greater_col_idx

            elif isinstance(arg, StrictLessThan) or isinstance(arg, LessThan):

                # Following Propositional logic, a predicate (a => r < value) can be translated as: r - value < 0
                # Alternatively, flux predicate a => r<value can be a(value + tolerance - r_LB) + r > value + tolerance
                # & a(r_UB - value - tolerance) + r < r_UB

                # So, for the mid term expression a => r > value
                # We have therefore the equations: a(value + tolerance - r_LB) + r > value + tolerance
                # & a(r_UB - value - tolerance) + r < r_UB

                # In matlab: c * (const - Vmax) + v <= const
                # In matlab: Vmin <= c * (Vmin - const) + v

                # In matlab: compareVal = -compareVal + epsilon
                # In matlab: matrix_values = [compareVal-vmax, 1] (-inf, compareVal)
                # In matlab: matrix_values = [vmin-compareVal, 1] (vmin, inf)

                # add Less variable / column
                # less_col_idx = self.__add_column__('mid_term_' + str(self.A.shape[1]))
                less_col_idx = self.__add_column__('mid_term_' + str(len(self.A[0])))

                # Less left operators
                # op_l can be growth, ph, ...
                # otherwise, it can be a reaction or a boundary reaction associated with a metabolite
                # Either way, they should be already in the matrix

                if arg.args[0].is_Symbol:
                    op_l = traversal_results[arg.args[0]]
                    c_val = arg.args[1]

                else:
                    op_l = traversal_results[arg.args[1]]
                    c_val = arg.args[0]

                _lb, _ub = self.lbs[op_l], self.ubs[op_l]

                # add Less row (a(value + tolerance - r_LB) + r > value + tolerance) and set mip bounds to
                # -inf;-comparison_val
                self.__add_row__([(less_col_idx, c_val + _SRFBA_TOL - _lb), (op_l, 1)], a_lb=c_val + _SRFBA_TOL,
                                 a_ub=999999999)
                # self.__add_row__([(less_col_idx, -c_val - _ub), (op_l, 1)], a_lb=-999999, a_ub=-c_val)

                # add Less row (a(r_UB - value - tolerance) + r < r_UB) and set mip bounds to lb;inf
                self.__add_row__([(less_col_idx, _ub - c_val - _SRFBA_TOL), (op_l, 1)], a_lb=-999999999, a_ub=_ub)
                # self.__add_row__([(less_col_idx, _lb + c_val), (op_l, 1)], a_lb=_lb, a_ub=999999)

                traversal_results[arg] = less_col_idx
                last_index = less_col_idx

            else:

                # If the argument is not an And, Or, Not or Relational
                # it must be a symbol or a number from the flux predicates

                # If it is a symbol, it must be added to the matrix
                if arg.is_Symbol:

                    # However, some things should be handle before adding the variable:

                    # special case of ph
                    # ph can be treated as a reaction
                    # as other reactions, the lower and upper bounds can be set to the initial state or vary between
                    # 0 and 14
                    # the row bounds (a_lb and a_ub) must be equal to zero as it is done on the reactions
                    if arg.name.lower() == 'ph':

                        if arg.name in self.initial_state:
                            arg_idx = self.__add_column__(arg.name, self.initial_state[arg.name],
                                                          self.initial_state[arg.name], v_type='continuous')
                            traversal_results[arg] = arg_idx
                            last_index = arg_idx

                        else:
                            arg_idx = self.__add_column__(arg.name, 0, 14, v_type='continuous')
                            traversal_results[arg] = arg_idx
                            last_index = arg_idx

                    # special case of growth (always positive)
                    # growth can be treated as a reaction
                    # as other reactions, the lower and upper bounds can be set to the initial state or vary between
                    # 0.1 and infinite (always positive)
                    # the row bounds (a_lb and a_ub) must be equal to zero as it is done on the reactions
                    elif arg.name.lower() == 'growth':

                        if arg.name in self.initial_state:
                            arg_idx = self.__add_column__(arg.name, self.initial_state[arg.name],
                                                          self.initial_state[arg.name], v_type='continuous')
                            traversal_results[arg] = arg_idx
                            last_index = arg_idx

                        else:
                            arg_idx = self.__add_column__(arg.name, 0 + _SRFBA_TOL, 999999, v_type='continuous')
                            traversal_results[arg] = arg_idx
                            last_index = arg_idx

                    else:

                        var_id = arg.name

                        if var_id in self._aliases_map:
                            var_id = self._aliases_map[var_id]

                        if var_id in self.metabolic_regulatory_reactions:

                            # if the arg is a reaction, the real reaction index should be returned
                            reaction = self._metabolic_regulatory_reactions[var_id].cbm_model_id

                            rxn_idx = self.__add_column__(reaction)
                            traversal_results[arg] = rxn_idx
                            last_index = rxn_idx

                        elif var_id in self.metabolic_regulatory_metabolites:

                            # if the arg is a metabolite, the associated reaction index should be returned

                            reaction = self.cbm_simulation_interface.get_boundary_reaction(
                                self._metabolic_regulatory_metabolites[arg.name].cbm_model_id)

                            if not reaction:
                                # The metabolite does not have a reaction associated with it. This is bad sign though
                                reaction = var_id

                            rxn_idx = self.__add_column__(reaction)
                            traversal_results[arg] = rxn_idx
                            last_index = rxn_idx

                        elif var_id in self.targets:
                            # if the arg is in targets only the column is added, as in the next iterations the row will be
                            # created

                            tg_idx = self.__add_column__(var_id)
                            traversal_results[arg] = tg_idx
                            last_index = tg_idx

                        else:

                            # if the args of a sympy's expression are regulators or genes,
                            # which are not defined in self.targets or self.reactions, they will never be created elsewhere
                            # they must be added to the MILP matrix as columns
                            # The state can be inferred from the initial state if the variables are in the initial state
                            # else, the variables are considered as unknown

                            if var_id in self.initial_state:

                                arg_idx = self.__add_column__(var_id, self.initial_state[var_id],
                                                              self.initial_state[var_id])
                                traversal_results[arg] = arg_idx
                                last_index = arg_idx

                            else:
                                arg_idx = self.__add_column__(var_id)
                                traversal_results[arg] = arg_idx
                                last_index = arg_idx

        # add gene row which means that the gene variable in the mip matrix is associated with the last mid term
        # expression, namely the whole expression
        # set mip bounds to 0;0
        self.__add_row__([(variable, 1), (last_index, -1)], a_ub=0)
        return

    def build(self):

        """
        Method for building SRFBA's MILP problem. It must be ran whenever a change or set of changes (adding a
        regulator, regulatory interaction, ...) is made to the model!
        Otherwise, the latest changes won't be considered for the simulation.

        Simulate method builds the problem when the flag build is used.

        The problem includes:
            - matrix A (rows are odes and columns are variables)
            - ode's lower and upper bounds
            - continuous and boolean variables
            - variables' lower and upper bounds

        A boolean variable is created for each target, regulator, regulatory condition (e.g. pH), metabolic gene,
        reaction and mid-term operators available in the regulatory interactions and GPRs.
        For instance:

        PFL: pyruvate + coa -> formate + acetyl-coa

        R_PFL, G1

        A = [[-1, 0, 0], # (0,0)
            [ -1, 0, 0], # (0,0)
            [  1, 0, 0], # (0,0)
            [  1, 0, 0], # (0,0)
            [  0, 1, -1]] # (0,0)

        lbs, ubs = (0,0,0), (1000, 1, 1)

        The MILP problem is then passed to the cobamp engine for simulating generic linear systems.

        :return:
        """

        # cbm_model S matrix (metabolites, reactions)
        self._A = self.cbm_simulation_interface.get_S()

        self._lbs, self._ubs = self.cbm_simulation_interface.get_bounds()

        self._variables = {r: i for i, r in enumerate(self.reactions)}
        self._variables_types['continuous'] = list(self.variables.values())

        # using list plus append results into significant speed improvements
        self._A = [list(row) for row in self.A]
        self._lbs, self._ubs = list(self.lbs), list(self.ubs)

        self._a_lb = [0] * len(self.A)
        self._a_ub = [0] * len(self.A)

        # regulatory rules over genes
        for reg_inter in self.regulatory_interactions_gen():

            # add target boolean variable / add column / Add MILP bool variable
            # If already exists, only the index is returned
            target_bool_idx = self.__add_column__(reg_inter.target.id)

            # If the expression is empty, it means that the target or reaction can take any boolean value (0;1)
            # add row and set mip bounds to 0;1
            if reg_inter.sympify is None:

                # The initial state can have however the state of this variable
                if reg_inter.target.id in self.initial_state:
                    self.__add_row__([(target_bool_idx, 1)], a_lb=self.initial_state[reg_inter.target.id],
                                     a_ub=self.initial_state[reg_inter.target.id])

                else:

                    self.__add_row__([(target_bool_idx, 1)])

            else:

                self.__boolean_rule_decode__(target_bool_idx, reg_inter.sympify)

        # gprs over reactions
        for i, (rxn, (_, _, sympify, _, _, _)) in enumerate(self._gprs_evaluator.items()):

            # add reaction boolean variable / add column / Add MILP bool variable
            # If already exists, only the index is returned
            reaction_bool_idx = self.__add_column__('B_' + rxn)

            # If the expression is empty, it means that the target or reaction can take any boolean value (0;1)
            # add row and set mip bounds to 0;1
            if sympify is None:

                if rxn in self._aliases_map:
                    rxn = self._aliases_map[rxn]

                # The initial state can have however the state of this variable
                if rxn in self.initial_state:
                    self.__add_row__([(reaction_bool_idx, 1)], a_lb=self.initial_state[rxn],
                                     a_ub=self.initial_state[rxn])

                else:

                    self.__add_row__([(reaction_bool_idx, 1)])

            else:

                self.__boolean_rule_decode__(reaction_bool_idx, sympify)

            self.__decode_reaction_bounds__(i, reaction_bool_idx)

        self._A = np.array(self.A)

        self._problem = GenericLinearSystem(self.A,
                                            VAR_CONTINUOUS,
                                            self.lbs,
                                            self.ubs,
                                            self.a_lb,
                                            self.a_ub,
                                            list(self.variables.keys()))
        # build cobamp problem
        self.problem.build_problem()

        # var types can be continuous or binary, but only the variables after the reactions are binary
        self.problem.set_variable_types([self.problem.model.variables[i] for i in self._variables_types['continuous']],
                                        VAR_CONTINUOUS)
        self.problem.set_variable_types([self.problem.model.variables[i] for i in self._variables_types['boolean']],
                                        VAR_BINARY)

    def simulate(self, build=False, objective=None, maximize=True):

        """

        SRFBA model simulation. Simulation is performed using cobamp's linear system optimizer.

        See build method for further insight over SRFBA's MILP problem and how it is solved.

        :param build: bool, build the milp problem?. False is the default
        :param objective: None or dict, objective for the optimization. None for the current objective
        :param maximize: bool, direction for the optimization.
        :return: SimulationResult object, it contains the metabolic and regulatory states too.
        
        Solutions are also stored under solution property
        """

        if build:
            self.build()

        if not self.A:
            self.build()

        self.objective = objective
        self.maximize = maximize

        lso = LinearSystemOptimizer(self.problem, build=False)

        # objective function
        objective_ids = self.cbm_simulation_interface.get_objective()
        c = np.zeros(self.A.shape[1])
        for var in objective_ids:
            c[self.variables[var]] = 1
        minimize = not self.maximize

        self.problem.set_objective(c, minimize)

        solution = lso.optimize()

        #building mewpy's solution object
        flxs = {}
        bools = {}
        regulatory_conditions = {}

        for key, val in solution.var_values().items():

            if key in self.reactions:
                flxs[key] = val
            elif key in self.regulatory_conditions:
                regulatory_conditions[key] = val
            else:
                bools[key] = val

        solution = SimulationResult(self.cbm_model,
                                    solution.objective_value(),
                                    fluxes=flxs,
                                    status=solution.status(),
                                    envcond=self.environmental_conditions,
                                    model_constraints=self.cbm_simulation_interface.constraints,
                                    simul_constraints={},
                                    maximize=maximize)

        solution.regulatory_solution = bools
        solution.regulatory_conditions = regulatory_conditions

        self._regulatory_solution = bools.update(regulatory_conditions)
        self._solution = solution

        return self.solution


"""JMetalpy operators
"""

from mewpy.optimization.jmetal.problem import KOSolution, OUSolution
from mewpy.utils.constants import EAConstants
from jmetal.core.operator import Mutation, Crossover
from jmetal.core.solution import Solution
from typing import List
import random
import copy


class ShrinkMutation(Mutation[Solution]):
    """ Shrink mutation. A gene is removed from the solution.

    :param probability: (float), The mutation probability.
    :param min_size: (int) the solution minimum size.
    
    """
    def __init__(self, probability: float = 0.1,  min_size :int = EAConstants.MIN_SOLUTION_SIZE):
        super(ShrinkMutation, self).__init__(probability = probability)
        self.min_size = min_size

    def execute(self, solution: Solution) -> Solution:
        """
        Apply the mutation.
        
        :param solution: The candidate solution to be mutated.
        :returns: A mutated solution.
        
        """
        if random.random() <= self.probability and solution.number_of_variables > self.min_size:
            var = copy.copy(solution.variables)
            index = random.randint(0, len(var) - 1)
            del var[index]
            solution.variables = var
            solution.number_of_variables = len(var) 
        return solution

    def get_name(self):
        return 'Shrink Mutation'



class GrowMutationKO(Mutation[KOSolution]):
    """ Grow mutation. A gene is added to the solution.
 
    :param probability: (float), The mutation probability.
    :param min_size: (int) the solution minimum size.
    
    """

    def __init__(self, probability: float = 0.1,  max_size :int = EAConstants.MAX_SOLUTION_SIZE):
        super(GrowMutationKO, self).__init__(probability = probability)
        self.max_size = max_size


    def execute(self, solution: Solution) -> Solution:
        """
        Apply the mutation.
        
        :param solution: The candidate solution to be mutated.
        :returns: A mutated solution.
        
        """
        if random.random() <= self.probability and solution.number_of_variables < self.max_size:
            mutant = copy.copy(solution.variables)
            newElem = random.randint(solution.lower_bound,solution.upper_bound)
            while newElem in mutant:
                newElem = random.randint(solution.lower_bound,solution.upper_bound)
            mutant.append(newElem)
            solution.variables = mutant
            solution.number_of_variables = len(mutant)
        return solution



    def get_name(self):
        return 'Grow Mutation KO'



class GrowMutationOU(Mutation[OUSolution]):
    """ Grow mutation. A gene is added to the solution.
 
    :param probability: (float), The mutation probability.
    :param min_size: (int) the solution minimum size.
    
    """

    def __init__(self, probability: float = 0.1,  max_size :int = EAConstants.MAX_SOLUTION_SIZE):
        super(GrowMutationOU, self).__init__(probability = probability)
        self.max_size = max_size


    def execute(self, solution: Solution) -> Solution:
        """
        Apply the mutation.
        
        :param solution: The candidate solution to be mutated.
        :returns: A mutated solution.
        
        """
        if random.random() <= self.probability and solution.number_of_variables < self.max_size:
            mutant = copy.copy(solution.variables)
            idx = random.randint(solution.lower_bound[0],solution.upper_bound[0])
            idxs = [a for (a,b) in mutant]
            while idx in idxs:
                idx = random.randint(solution.lower_bound[0],solution.upper_bound[0])
            lv = random.randint(solution.lower_bound[1],solution.upper_bound[1])    
            mutant.append((idx,lv))
            solution.variables = mutant
            solution.number_of_variables = len(mutant)
        return solution



    def get_name(self):
        return 'Grow Mutation OU'



class UniformCrossoverKO(Crossover[KOSolution,KOSolution]):
    """Uniform Crossover for KO solutions

    :param probability: (float) The probability of crossover.
    :param max_size: (int) The solution maximum size.
    
    """
    def __init__(self, probability: float = 0.1, max_size: int = EAConstants.MAX_SOLUTION_SIZE):
        super(UniformCrossoverKO, self).__init__(probability=probability)
        self.max_size = max_size

    def execute(self, parents: List[KOSolution]) -> List[KOSolution]:
        if len(parents) != 2:
            raise Exception('The number of parents is not two: {}'.format(len(parents)))

        offspring = [copy.deepcopy(parents[0]), copy.deepcopy(parents[1])]

        if random.random() <= self.probability and  (offspring[0].number_of_variables >1 or offspring[1].number_of_variables >1):
            mom = set(copy.copy(offspring[0].variables))
            dad = set(copy.copy(offspring[1].variables))
            intersection = mom & dad
            otherElems = list((mom | dad).difference(intersection))
            child1 = copy.copy(intersection)
            child2 = copy.copy(intersection)

            while len(otherElems) > 0:
                elemPosition = random.randint(0, len(otherElems) - 1) if len(otherElems)>1 else 0
                if len(child1) == self.max_size or len(child2) == 0:
                    child2.add(otherElems[elemPosition])
                elif len(child2) == self.max_size or len(child1) == 0:
                    child1.add(otherElems[elemPosition])
                else:
                    r = random.random()
                    if r <= 0.5:
                        child1.add(otherElems[elemPosition])
                    else:
                        child2.add(otherElems[elemPosition])
                otherElems.pop(elemPosition)
            
            offspring[0].variables=list(child1)
            offspring[0].number_of_variables = len(child1)
            offspring[1].variables=list(child2)
            offspring[1].number_of_variables = len(child2)
        return offspring



    def get_number_of_parents(self) -> int:
        return 2

    def get_number_of_children(self) -> int:
        return 2

    def get_name(self):
        return 'Uniform Crossover KO'



class MutationContainer(Mutation[Solution]):
    """A container for the mutation operators.
    
    :param probability: (float) The probability of applying a mutation.
    :param mutators: (list) The list of mutators.
    
    """
    
    def __init__(self,probability:float = 0.5, mutators = []):
        super(MutationContainer, self).__init__(probability=probability)
        self.mutators = mutators


    def execute(self, solution: Solution) -> Solution:
        # randomly select a mutator and apply it 
        if random.random() <= self.probability:
            idx = random.randint(0,len(self.mutators)-1)
            mutator = self.mutators[idx]
            return mutator.execute(solution)
        else:
            return solution


    def get_name(self):
        return 'Mutation container'
    

class UniformCrossoverOU(Crossover[OUSolution,OUSolution]):
    """
        Uniform Crossover for OU solutions
    """
    def __init__(self, probability: float = 0.1, max_size: int = EAConstants.MAX_SOLUTION_SIZE):
        super(UniformCrossoverOU, self).__init__(probability=probability)
        self.max_size = max_size

    def execute(self, parents: List[KOSolution]) -> List[KOSolution]:
        if len(parents) != 2:
            raise Exception('The number of parents is not two: {}'.format(len(parents)))

        offspring = [copy.deepcopy(parents[0]), copy.deepcopy(parents[1])]

        if random.random() <= self.probability and  (offspring[0].number_of_variables >1 or offspring[1].number_of_variables >1):
            mom = set(copy.copy(offspring[0].variables))
            dad = set(copy.copy(offspring[1].variables))
            intersection = mom & dad
            otherElems = list((mom | dad).difference(intersection))
            child1 = copy.copy(intersection)
            child2 = copy.copy(intersection)

            while len(otherElems) > 0:
                elemPosition = random.randint(0, len(otherElems) - 1) if len(otherElems)>1 else 0
                if len(child1) == self.max_size or len(child2) == 0:
                    child2.add(otherElems[elemPosition])
                elif len(child2) == self.max_size or len(child1) == 0:
                    child1.add(otherElems[elemPosition])
                else:
                    r = random.random()
                    if r <= 0.5:
                        child1.add(otherElems[elemPosition])
                    else:
                        child2.add(otherElems[elemPosition])
                otherElems.pop(elemPosition)
            
            offspring[0].variables=list(child1)
            offspring[0].number_of_variables = len(child1)
            offspring[1].variables=list(child2)
            offspring[1].number_of_variables = len(child2)
        return offspring



    def get_number_of_parents(self) -> int:
        return 2

    def get_number_of_children(self) -> int:
        return 2

    def get_name(self):
        return 'Uniform Crossover OU'



class SingleMutationKO(Mutation[KOSolution]):
    """
    Mutates a single element
    """
    def __init__(self, probability: float = 0.1):
        super(SingleMutationKO, self).__init__(probability = probability)
        

    def execute(self, solution: Solution) -> Solution:
        if random.random() <= self.probability:
            mutant = copy.copy(solution.variables)
            index = random.randint(0, len(mutant) - 1)
            newElem = random.randint(solution.lower_bound,solution.upper_bound)
            while newElem in mutant:
                newElem = random.randint(solution.lower_bound,solution.upper_bound)
            mutant[index] = newElem
            solution.variables = mutant
        return solution



    def get_name(self):
        return 'Single Mutation KO'



class SingleMutationOU(Mutation[OUSolution]):
    """
    Mutates a single element
    """
    def __init__(self, probability: float = 0.1):
        super(SingleMutationOU, self).__init__(probability = probability)
        
    def execute(self, solution: Solution) -> Solution:
        if random.random() <= self.probability:
            mutant = copy.copy(solution.variables)
            l = [i for (i,j) in mutant]
            index = random.randint(0, len(mutant) - 1)
            idx, idy = mutant[index] 
            is_mutate_idx = False
            if random.random() > 0.5:
                idx = random.randint(solution.lower_bound[0],solution.upper_bound[0])
                while idx in l:
                    idx = random.randint(solution.lower_bound[0],solution.upper_bound[0])
                is_mutate_idx = True
            lv = random.randint(solution.lower_bound[1],solution.upper_bound[1])
            if not is_mutate_idx:
                while lv == idy:
                    lv = random.randint(solution.lower_bound[1],solution.upper_bound[1])
            mutant[index] = (idx,lv)
            solution.variables = mutant
        return solution

    def get_name(self):
        return 'Single Mutation KO'

    
class SingleMutationOULevel(Mutation[OUSolution]):
    """
    Mutates the expression level of a single element
    """
    def __init__(self, probability: float = 0.1):
        super(SingleMutationOULevel, self).__init__(probability = probability)
        
    def execute(self, solution: Solution) -> Solution:
        if random.random() <= self.probability:
            mutant = copy.copy(solution.variables)
            index = random.randint(0, len(mutant) - 1)
            idx, idy = mutant[index] 
            lv = random.randint(solution.lower_bound[1],solution.upper_bound[1])
            while lv == idy:
                lv = random.randint(solution.lower_bound[1],solution.upper_bound[1])
            mutant[index] = (idx,lv)
            solution.variables = mutant
        return solution



    def get_name(self):
        return 'Single Mutation KO'





from mewpy.utils.ode import KineticConfigurations, SolverConfigurations, ODEStatus, ODESpySolver
from mewpy.simulation.simulation import SimulationResult
from mewpy.model.kinetic import ODEModel
from collections import OrderedDict
import threading
import warnings


def kinetic_solve(model, finalParameters, finalFactors, initialConc, timePoints):
    """
    Private function: auxiliary function required to avoid the pickling the solver.solve function

    """
    finalRates = OrderedDict()
    f = model.get_ode(r_dict=finalRates, params=finalParameters, factors=finalFactors)

    def func(x, t):
        return f(t, x)

    solver = ODESpySolver(KineticConfigurations.SOLVER_METHOD).get_solver(func)
    solver.set_initial_condition(list(initialConc.values()))
    try:
        X, _ = solver.solve(timePoints)
        # if solver returns a solution where any concentration is negative
        for c in X[1]:
            if c < -1*SolverConfigurations.RELATIVE_TOL:
                return ODEStatus.ERROR, {}, {}

    except Exception as e:
        warnings.warn(f"ODE Solver error: {e}.")
        return ODEStatus.ERROR, {}, {}

    # values bellow solver precision will be set to 0
    finalRates.update({k: 0 for k, v in finalRates.items() if
                       v < SolverConfigurations.ABSOLUTE_TOL and v > - SolverConfigurations.ABSOLUTE_TOL})
    conc = OrderedDict(zip(model.metabolites.keys(), X[1]))
    return ODEStatus.OPTIMAL, finalRates, conc


class KineticThread(threading.Thread):
    """
    Solves the ODE inside a thread enabling to impose a timeout limit with thread.join(timeout)

    """
    def __init__(self, model, parameters=None, final_factors=None, initial_concentrations=None, time_steps=None):
        super(KineticThread, self).__init__()
        self.model = model
        self.parameters = parameters
        self.final_factors = final_factors
        self.initial_concentrations = initial_concentrations
        self.time_steps = time_steps
        self.status = None
        self.rates = None
        self.concentrations = None

    def run(self):
        try:
            self.status, self.rates, self.concentrations = kinetic_solve(self.model, self.parameters,
                                                                         self.final_factors,
                                                                         self.initial_concentrations,
                                                                         self.time_steps)
        except Exception as e:
            warnings.warn(f"{e}")
        return


class kineticSimulationResult(SimulationResult):
    def __init__(self, model, status, factors=None, rates=None, concentrations=None):
        super(kineticSimulationResult, self).__init__(model, None, fluxes=rates, status=status)
        self.factors = factors
        self.concentations = concentrations


class KineticSimulation:

    def __init__(self, model, parameters=None, tSteps=[0, 1e9], timeout=KineticConfigurations.SOLVER_TIMEOUT,
                 solver=KineticConfigurations.SOLVER, method=KineticConfigurations.SOLVER_METHOD):
        if not isinstance(model, ODEModel):
            raise ValueError('model is not an instance of ODEModel.')
        self.model = model
        self.parameters = parameters
        self.tSteps = tSteps
        self.timeout = timeout
        self.solver = solver
        self.method = method

    def __getstate__(self):
        state = OrderedDict(self.__dict__.copy())
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)

    def get_initial_concentrations(self):
        return self.model.concentrations

    def get_time_steps(self):
        return self.tSteps

    def simulate(self, factors=None):
        """
        This method preform the phenotype simulation of the kinetic model, using the solverId method and applying
        the modifications present in the instance of overrideSimulProblem.

        :param dict factores: Modification over the kinetic model.
        :returns: Returns a kineticSimulationResult with the steady-state flux distribution and concentrations.

        """

        final_factors = factors if factors is not None else {}
        # update initial concentrations when a [enz] is changed: == 0, up or down regulated
        initConcentrations = self.get_initial_concentrations().copy()

        status = None
        sstateRates = None
        sstateConc = None

        if self.timeout:

            th = KineticThread(self.model, parameters=self.parameters,
                               final_factors=final_factors,
                               initial_concentrations=initConcentrations,
                               time_steps=self.get_time_steps())

            th.start()
            th.join(self.timeout)
            status = th.status
            sstateRates = th.rates
            sstateConc = th.concentrations

        else:
            status, sstateRates, sstateConc = kinetic_solve(self.model, self.parameters,
                                                            final_factors,
                                                            initConcentrations,
                                                            self.get_time_steps())

        return kineticSimulationResult(self.model, status, factors=final_factors, rates=sstateRates,
                                       concentrations=sstateConc)



from .problem import AbstractKOProblem, AbstractOUProblem
from mewpy.utils.constants import ModelConstants
from collections import OrderedDict
import warnings


class GeckoKOProblem(AbstractKOProblem):
    """Gecko KnockOut Optimization Problem

    :param model: The constraint based metabolic model.
    :param list fevaluation: a list of callable EvaluationFunctions.

    Optional:

    :param OrderedDict envcond: environmental conditions.
    :param OrderedDict constraints: additional constraints to be applied to the model.
    :param int candidate_min_size: The candidates minimum size.
    :param int candidate_max_size: The candidates maximum size.
    :param list target: List of target reactions.
    :param list non_target: List of non target reactions. Not considered if a target list is provided.
    :param float scalefactor: a scaling factor to be used in the LP formulation.
    :param str prot_prefix: the protein draw reaction prefix. Default `draw_prot_`.


    Note:

    Targets as well as non target proteins are defined using their prot id, ex  `P0351`, and not by the associated draw
    reaction id, ex `draw_prot_P0351`.

    """

    def __init__(self, model, fevaluation=None, **kwargs):
        # if isinstance(model,GeckoModel):
        super(GeckoKOProblem, self).__init__(
            model, fevaluation=fevaluation, **kwargs)
        # else:
        #    raise Exception("The model should be an instance of GeckoModel")
        # problem simulation context
        self.prot_prefix = kwargs.get('prot_prefix', 'draw_prot_')

    def _build_target_list(self):
        """
        If not provided, targets are all non essential proteins.
        """
        proteins = set(self.simulator.proteins)
        # as draw_prot_XXXXXX
        ess = self.simulator.essential_proteins(self.prot_prefix)
        # remove 'draw_prot_'
        n = len(self.prot_prefix)
        essential = set([p[n:] for p in ess])
        target = proteins - essential
        if self.non_target:
            target = target - set(self.non_target)
        self._trg_list = list(target)

    def decode(self, candidate):
        """
        Decodes a candidate, an integer set, into a dictionary of constraints
        """
        constraints = OrderedDict()
        for idx in candidate:
            try:
                constraints["{}{}".format(
                    self.prot_prefix, self.target_list[idx])] = 0
            except IndexError:
                raise IndexError(
                    f"Index out of range: {idx} from {len(self.target_list[idx])}")
        return constraints


class GeckoOUProblem(AbstractOUProblem):
    """
    Gecko Under/Over expression Optimization Problem


    :param model (metabolic model): The constraint based metabolic model.
    :param fevaluation (list): a list of callable EvaluationFunctions.

    Optional:

    :param OrderedDict envcond: environmental conditions.
    :param OrderedDict constraints: additional constraints to be applied to the model.
    :param int candidate_min_size: The candidates minimum size.
    :param int candidate_max_size: The candidates maximum size.
    :param list target: List of target reactions.
    :param list non_target: List of non target reactions. Not considered if a target list is provided.
    :param float scalefactor: a scaling factor to be used in the LP formulation.
    :param dic reference: Dictionary of flux values to be used in the over/under expression values computation.
    :param str prot_prefix: the protein draw reaction prefix. Default `draw_prot_`.


    Note:
    Target as well as non target proteins are defined with their prot id, ex `P0351`, and with the associated reaction
    id, ex `draw_prot_P0351`.

    """

    def __init__(self, model, fevaluation=None, **kwargs):
        # if isinstance(model,GeckoModel):
        super(GeckoOUProblem, self).__init__(
            model, fevaluation=fevaluation, **kwargs)
        # else:
        #    raise Exception("The model should be an instance of GeckoModel")
        # problem simulation context
        self.prot_rev_reactions = None
        self.prot_prefix = kwargs.get('prot_prefix', 'draw_prot_')

    def _build_target_list(self):
        """
        If not provided, targets are all non essential proteins.
        """
        proteins = set(self.simulator.proteins)
        target = proteins
        if self.non_target:
            target = target - set(self.non_target)
        self._trg_list = list(target)

    def decode(self, candidate):
        """
        Decodes a candidate, a set (idx,lv), into a dictionary of constraints
        Reverseble reactions associated to proteins with over expression are KO
        according to the flux volume in the wild type.

        :param candidate: The candidate to be decoded.
        :returns: A dictionary of metabolic constraints.
        """
        constraints = OrderedDict()

        if self.prot_rev_reactions is None:
            self.prot_rev_reactions = self.simulator.protein_rev_reactions

        for idx, lv_idx in candidate:
            try:
                prot = self.target_list[idx]
                rxn = self.prot_prefix+prot
                lv = self.levels[lv_idx]
                fluxe_wt = self.reference[rxn]
                if lv < 0:
                    raise ValueError("All UO levels should be positive")
                # a level = 0 is interpreted as KO
                elif lv == 0:
                    constraints[rxn] = 0.0
                # under expression
                elif lv < 1:
                    constraints[rxn] = (0.0, lv * fluxe_wt)
                # TODO: Define how a level 1 is tranlated into constraints...
                elif lv == 1:
                    continue
                else:
                    constraints[rxn] = (
                        lv*fluxe_wt, ModelConstants.REACTION_UPPER_BOUND)
                    # Deals with reverse reactions associated with the protein.
                    # Strategy: The reaction direction with no flux in the wild type (reference) is KO.
                    if prot in self.prot_rev_reactions.keys():
                        reactions = self.prot_rev_reactions[prot]
                        for r, r_rev in reactions:
                            if self.reference[r] == 0 and self.reference[r_rev] == 0:
                                continue
                            elif self.reference[r] > 0 and self.reference[r_rev] == 0:
                                constraints[r_rev] = 0.0
                            elif self.reference[r] == 0 and self.reference[r_rev] > 0:
                                constraints[r] = 0.0
                            else:
                                warnings.warn(
                                    f"Reactions {r} and {r_rev}, associated with the protein {prot},\
                                    both have fluxes in the WT.")
            except IndexError:
                raise IndexError("Index out of range")

        return constraints



from jmetal.core.observer import Observer
from jmetal.lab.visualization import StreamingPlot
from mewpy.optimization.ea import non_dominated_population
from typing import List, TypeVar
import logging
import numpy
import copy

S = TypeVar('S')
LOGGER = logging.getLogger('mewpy')


class VisualizerObserver(Observer):

    def __init__(self,
                 reference_front: List[S] = None,
                 reference_point: list = None,
                 display_frequency: float = 1.0,
                 non_dominated=True) -> None:
        self.figure = None
        self.display_frequency = display_frequency
        self.reference_point = reference_point
        self.reference_front = reference_front
        self.non_dominated = non_dominated

    def update(self, *args, **kwargs):
        evaluations = kwargs['EVALUATIONS']
        solutions = kwargs['SOLUTIONS']

        if solutions:
            if self.figure is None:

                axis_labels = None
                problem = kwargs['PROBLEM']
                if problem and problem.obj_labels:
                    axis_labels = problem.obj_labels

                self.figure = StreamingPlot(reference_point=self.reference_point,
                                            reference_front=self.reference_front,
                                            axis_labels=axis_labels)
                self.figure.plot(solutions)

            if (evaluations % self.display_frequency) == 0:
                # check if reference point has changed
                reference_point = kwargs.get('REFERENCE_POINT', None)
                # negative fitness values are converted to positive
                population = copy.copy(solutions)
                if self.non_dominated:
                    population = non_dominated_population(population)
                for i in range(len(population)):
                    obj = [abs(x) for x in population[i].objectives]
                    population[i].objectives = obj

                if reference_point:
                    self.reference_point = reference_point
                    self.figure.update(population, reference_point)
                else:
                    self.figure.update(population)

                self.figure.ax.set_title(
                    'Eval: {}'.format(evaluations), fontsize=13)


class PrintObjectivesStatObserver(Observer):

    def __init__(self, frequency: float = 1.0) -> None:
        """ Show the number of evaluations, best fitness and computing time.

        :param frequency: Display frequency. """
        self.display_frequency = frequency
        self.first = True

    def fitness_statistics(self, solutions):
        """Return the basic statistics of the population's fitness values.       
        Arguments:roblem = kwargs['PROBLEM']
        """

        stats = {}
        first = solutions[0].objectives
        n = len(first)
        for i in range(n):
            f = [abs(p.objectives[i]) for p in solutions]
            worst_fit = min(f)
            best_fit = max(f)
            med_fit = numpy.median(f)
            avg_fit = numpy.mean(f)
            std_fit = numpy.std(f)
            stats['obj_{}'.format(i)] = {'best': best_fit, 'worst': worst_fit,
                                         'mean': avg_fit, 'median': med_fit, 'std': std_fit}
        return stats

    def stats_to_str(self, stats, evaluations, title=False):
        if title:
            title = "Eval(s)|"
        values = " {0:>6}|".format(evaluations)

        for key in stats:
            s = stats[key]
            if title:
                title = title + "     Worst      Best    Median   Average   Std Dev|"
            values = values + "  {0:.6f}  {1:.6f}  {2:.6f}  {3:.6f}  {4:.6f}|".format(s['worst'],
                                                                                      s['best'],
                                                                                      s['median'],
                                                                                      s['mean'],
                                                                                      s['std'])
        if title:
            return title+"\n"+values
        else:
            return values

    def update(self, *args, **kwargs):
        evaluations = kwargs['EVALUATIONS']
        solutions = kwargs['SOLUTIONS']
        if (evaluations % self.display_frequency) == 0 and solutions:
            if type(solutions) == list:
                stats = self.fitness_statistics(solutions)
                message = self.stats_to_str(stats, evaluations, self.first)
                self.first = False
            else:
                fitness = solutions.objectives
                res = abs(fitness[0])
                message = 'Evaluations: {}\tFitness: {}'.format(
                    evaluations, res)
            print(message)



from enum import Enum

# Model specific simulators mapping:
# Entries take the form:  full_model_class_path -> (simulator_path, simulator_class_name)
# TODO: use qualified names

map_model_simulator = {
    'geckopy.gecko.GeckoModel': ('mewpy.simulation.cobra', 'GeckoSimulation'),
    'mewpy.model.gecko.GeckoModel': ('mewpy.simulation.reframed', 'GeckoSimulation'),
    'mewpy.model.smoment.SMomentModel': ('mewpy.simulation.reframed', 'GeckoSimulation')
}


def get_simulator(model, envcond=None, constraints=None, reference=None):
    """
    Returns a simulator instance for the model.
    The simulator instance is model dependent.
    Besides able to be used on its own, this function is invoked by EA optimization problems 
    and by evaluation function instances to perform phenotyoe evaluations of candidate solutions.

    
    :param model: the model
    :param dict envcond: A dictionary of environmental conditions.
    :param dict contrainsts: A dictionary of additional persistent constraints.
    :returns: An instance of Simulator
    """
    instance = None
    name = f"{model.__class__.__module__}.{model.__class__.__name__}"
    if name in map_model_simulator:
        module_name, class_name = map_model_simulator[name]
        module = __import__(module_name, fromlist=[None])
        class_ = getattr(module, class_name)
        instance = class_(model, envcond=envcond,
                          constraints=constraints, reference=reference)
    else:
        try:
            from cobra.core.model import Model
            if isinstance(model, Model):
                from .cobra import Simulation
                instance = Simulation(
                    model, envcond=envcond, constraints=constraints, reference=reference)
        except ImportError:
            pass
        if not instance:
            try:
                from reframed.core.cbmodel import CBModel
                if isinstance(model, CBModel):
                    from .reframed import Simulation
                    instance = Simulation(
                        model, envcond=envcond, constraints=constraints, reference=reference)
            except ImportError:
                pass
    if not instance:
        raise ValueError(f"The model <{name}> has no defined simulator.")
    return instance


def get_container(model):
    """
    Returns a container for a given model sharing a common interface.
    A container does not perform any task, it only serves as a basic interface with a phenotype simulator.

    :param model: An instance of a metabolic model.
    :returns: A container.
     
    """
    try:
        from reframed.core.cbmodel import CBModel
        if isinstance(model, CBModel):
            from mewpy.simulation.reframed import CBModelContainer
            return CBModelContainer(model)
    except:
        pass

    try:
        from cobra.core.model import Model
        if isinstance(model, Model):
            from mewpy.simulation.cobra import CobraModelContainer
            return CobraModelContainer(model)
    except:
        pass

    raise ValueError(f"Unrecognized model class: {model.__class__.name}")


class SimulationMethod(Enum):
    
    FBA   = 'FBA'
    pFBA  = 'pFBA'
    MOMA  = 'MOMA'
    lMOMA = 'lMOMA'
    ROOM  = 'ROOM'
    NONE  = 'NONE'
    
    def __eq__(self, other):
        """Overrides equal to enable string name comparison.
        Allows to seamlessly use: 
            SimulationMethod.FBA = SimulationMethod.FBA
            SimulationMethod.FBA = 'FBA'
        without requiring an additional level of comparison (SimulationMethod.FBA.name = 'FBA')
        """
        if isinstance(other,SimulationMethod):
            return super().__eq__(other)
        elif isinstance(other,str):
            return self.name == other
        else:
            return False

    def __hash__(self):
        return hash(self.name)


class SStatus(Enum):
    """ Enumeration of possible solution status. """
    OPTIMAL    = 'Optimal'
    UNKNOWN    = 'Unknown'
    SUBOPTIMAL = 'Suboptimal'
    UNBOUNDED  = 'Unbounded'
    INFEASIBLE = 'Infeasible'
    INF_OR_UNB = 'Infeasible or Unbounded'



from mewpy.simulation import get_simulator
import os


def load_reframed():

    DIR = os.path.dirname(os.path.realpath(__file__))
    PATH = os.path.join(DIR, '../../../examples/models/optram/')
    DATA_FILE = os.path.join(PATH, "yeast_7.6-optram.xml")

    from reframed.io.sbml import load_cbmodel
    model = load_cbmodel(DATA_FILE, flavor='cobra')
    model.summary()
    simul = get_simulator(model)
    # int(simul.reactions)


def load_ec_gecko():

    DIR = os.path.dirname(os.path.realpath(__file__))
    PATH = os.path.join(DIR, '../../../examples/models/autopacmen/')
    DATA_FILE = os.path.join(PATH, "iJO1366_2019_06_25_GECKO.xml")

    from mewpy.model.gecko import GeckoModel
    from reframed.io.sbml import load_cbmodel
    cbmodel = load_cbmodel(DATA_FILE)
    # ='R_ER_pool_TG_'
    model = GeckoModel(cbmodel, biomass_reaction_id='R_BIOMASS_Ec_iJO1366_WT_53p95M',
                       protein_reaction_id='R_PROTRS_TG_1', common_protein_pool_id='M_prot_pool')

    simul = get_simulator(model)
    for rxn in simul.reactions:
        if "R_PROT" in rxn:
            print(rxn)


def load_cobra():

    DIR = os.path.dirname(os.path.realpath(__file__))
    PATH = os.path.join(DIR, '../../../examples/models/optram/')
    DATA_FILE = os.path.join(PATH, "yeast_7.6-optram.xml")

    from cobra.io import read_sbml_model
    model = read_sbml_model(DATA_FILE)
    simul = get_simulator(model)
    simul.summary()


def load_gecko():
    from mewpy.model.gecko import GeckoModel
    model = GeckoModel('single-pool')
    simul = get_simulator(model)
    simul.summary()


if __name__ == "__main__":
    load_cobra()
    load_gecko()
    load_reframed()
    load_ec_gecko()



from reframed.core.cbmodel import CBModel
from reframed.external.cobrapy import to_cobrapy
import random
import string


def escher_maps():
    try:
        import escher
    except ImportError:
        raise RuntimeError("Escher is not installed.")

    maps = escher.list_available_maps()
    return [entry['map_name'] for entry in maps]


def randomString(stringLength=10):
    """Generate a random string of fixed length """
    letters = string.ascii_lowercase
    return ''.join(random.choice(letters) for i in range(stringLength))


def to_json(model, filename=None):
    import cobra
    if not filename:
        filename = randomString()+".json"
    if isinstance(model, cobra.core.model.Model):
        c_model = model
    elif isinstance(model, CBModel):
        c_model = to_cobrapy(model)
    else:
        raise Exception
    cobra.io.save_json_model(c_model, filename)
    return filename


def build_escher(model=None, fluxes=None, fmt_func=None, **kwargs):
    try:
        import escher
    except ImportError:
        raise RuntimeError("Escher is not installed.")

    js = None
    if isinstance(model, str) and model in escher_maps():
        map_name = model
    else:
        try:
            js = to_json(model)
        except Exception:
            map_name = 'e_coli_core.Core metabolism'

    if fluxes and fmt_func:
        fluxes = {fmt_func(r_id): val for r_id, val in fluxes.items()}
    else:
        fluxes = None
    if js:
        return escher.Builder(model_json=js, reaction_data=fluxes, **kwargs)
    else:
        return escher.Builder(map_name=map_name, reaction_data=fluxes, **kwargs)



import os
import pandas as pd
import numpy as np
from mewpy.utils.parsing import BOOLEAN_SPECIAL_CHARS


def read_tabular_regulatory_model(file_or_filepath, sep=None, id_col=None, rule_col=None, aliases_cols=None, header=None,
                                  special_chars=None, compartments_ids=None):

    if not sep:
        sep = ','

    if not id_col:
        id_col = 0

    if not rule_col:
        rule_col = 1

    if not isinstance(file_or_filepath, str):
        raise TypeError("File or file path must be a string")

    if not isinstance(sep, str):
        raise TypeError("Separator must be a string")

    if not isinstance(id_col, int):
        raise TypeError(
            "The column of genes identifiers must be its index in the file")

    if not isinstance(rule_col, int):
        raise TypeError(
            "The column of regulatory rules must be its index in the file")

    if aliases_cols:

        if not isinstance(aliases_cols, list):
            raise TypeError(
                "The aliases columns must be a list of all its indexes in the file")

        if not isinstance(aliases_cols[0], int):
            raise TypeError(
                "The aliases columns must be a list of all its indexes in the file")

    else:
        aliases_cols = []

    if not isinstance(header, int) and header is not None:
        raise TypeError("The header must be a int value or None")

    if special_chars:

        if not isinstance(special_chars, list):
            raise TypeError(
                "The special chars must be a list of all chars to be parsed out from the file")

        if not isinstance(special_chars[0], str):
            raise TypeError(
                "The special chars must be a list of all chars to be parsed out from the file")

        for char in special_chars:

            if char not in BOOLEAN_SPECIAL_CHARS:

                new_char = char.replace(')','_').replace('(','_').replace(' ','_')
                BOOLEAN_SPECIAL_CHARS[char] = '_' + new_char + '_'

    if compartments_ids:

        if not isinstance(compartments_ids, list):
            raise TypeError("The compartments ids must be a list of all compartments identifiers to be parsed out from "
                            "the file")

        if not isinstance(compartments_ids[0], str):
            raise TypeError("The compartments ids must be a list of all compartments identifiers to be parsed out from "
                            "the file")

        for char in special_chars:

            if char not in BOOLEAN_SPECIAL_CHARS:

                new_char = char.replace(')', '_').replace('(', '_').replace(' ', '_')
                BOOLEAN_SPECIAL_CHARS[char] = '_' + new_char + '_'

    names = {id_col: 'ids', rule_col: 'rules'}
    names.update({j: 'aliases_' + str(i + 1)
                  for i, j in enumerate(aliases_cols)})

    csv = pd.read_csv(file_or_filepath, sep, header=header)

    csv = csv.replace(np.nan, '', regex=True)

    cols = []

    for j, col in enumerate(csv.columns):
        if j in names:
            cols.append(names[j])
        else:
            del csv[col]

    csv.columns = cols
    csv.index = csv.loc[:, 'ids']

    return csv

def read_tabular_aliases(file_or_filepath, sep=None, id_col=None, aliases_cols=None, header=None):

    if not sep:
        sep = ','

    if not id_col:
        id_col = 0

    if not aliases_cols:
        aliases_cols = [1]

    if not isinstance(file_or_filepath, str):
        raise TypeError("File or file path must be a string")

    if not isinstance(sep, str):
        raise TypeError("Separator must be a string")

    if not isinstance(id_col, int):
        raise TypeError("The column of genes identifiers must be its index in the file")

    if not isinstance(aliases_cols, list):
        raise TypeError("The aliases columns must be a list of all its indexes in the file")

    if not isinstance(aliases_cols[0], int):
        raise TypeError("The aliases columns must be a list of all its indexes in the file")

    if not isinstance(header, int) and header is not None:
        raise TypeError("The header must be a int value or None")

    names = {id_col: 'ids'}
    names.update({j: 'aliases_' + str(i + 1) for i, j in enumerate(aliases_cols)})

    csv = pd.read_csv(file_or_filepath, sep, header=header)

    csv = csv.replace(np.nan, '', regex=True)

    cols = []

    for j, col in enumerate(csv.columns):
        if j in names:
            cols.append(names[j])
        else:
            del csv[col]

    csv.columns = cols
    csv.index = csv.loc[:, 'ids']

    return csv

if __name__ == '__main__':

    file = "iMC1010v4.csv"
    os.chdir('../../../examples/models/')

    df = read_tabular_regulatory_model(
        file, sep=';', id_col=2, rule_col=3, header=None, special_chars=['X'])
    print(df.columns)
    print(df.loc['XylE', 'rules'])
    print(type(df.loc['XylE', 'rules']))



"""
Simulation for REFRAMED models
"""

from reframed.cobra.simulation import FBA, pFBA, MOMA, lMOMA, ROOM
from reframed.solvers import solver_instance
from reframed.core.cbmodel import CBModel
from reframed.solvers.solution import Solution
from reframed.solvers.solution import Status as s_status
from mewpy.model.gecko import GeckoModel
from mewpy.simulation import SimulationMethod, SStatus
from mewpy.simulation.simulation import Simulator, SimulationResult, ModelContainer
from mewpy.utils.constants import ModelConstants
from mewpy.utils.parsing import evaluate_expression_tree
from collections import OrderedDict
import numpy as np
import logging


LOGGER = logging.getLogger(__name__)


class CBModelContainer(ModelContainer):
    """ A basic container for REFRAMED models.

    :param model: A metabolic model.

    """
    def __init__(self, model: CBModel):
        if not isinstance(model, CBModel):
            raise ValueError(
                "The model is not an instance of ReFramed CBModel")
        self.model = model

    @property
    def reactions(self):
        return list(self.model.reactions.keys())

    @property
    def genes(self):
        return list(self.model.genes.keys())

    @property
    def metabolites(self):
        return list(self.model.metabolites.keys())

    @property
    def compartments(self):
        return self.model.compartments

    def get_gpr(self, reaction_id):
        """Returns the gpr rule (str) for a given reaction ID.

        :param str reaction_id: The reaction identifier.
        :returns: A string representation of the GPR rule.

        """
        if reaction_id not in self.reactions:
            raise ValueError(f"Reactions {reaction_id} does not exist")
        reaction = self.model.reactions[reaction_id]
        if reaction.gpr:
            return str(reaction.gpr)
        else:
            return None

    def get_drains(self):
        return self.model.get_exchange_reactions()

    @property
    def medium(self):

        def is_active(rxn):
            """Determine if a boundary reaction permits flux towards creating
            metabolites
            """
            reaction = self.model.reactions[rxn]
            return ((bool(reaction.get_products()) and (reaction.ub > 0)) or
                    (bool(reaction.get_substrates) and (reaction.lb < 0)))

        def get_active_bound(rxn):
            """For an active boundary reaction, return the relevant bound"""
            reaction = self.model.reactions[rxn]
            if reaction.get_substrates():
                return -reaction.lb
            elif reaction.get_products():
                return reaction.ub

        return {rxn: get_active_bound(rxn) for rxn in self.get_drains()
                if is_active(rxn)}


class Simulation(CBModelContainer, Simulator):
    """Generic Simulation class for cobra Model.
       Defines the simulation conditions, and makes available a set of methods.

    :param model: An metabolic model instance.

    Optional:

    :param objective: The model objective.
    :param dic envcond: Dictionary of environmental conditions.
    :param dic constraints: A dictionary of reaction constraints.
    :param solver: An instance of the LP solver.
    :param dic reference: A dictionary of the wild type flux values.

    """

    def __init__(self, model: CBModel, objective=None, envcond=None, constraints=None,  solver=None, reference=None):

        if not isinstance(model, CBModel):
            raise ValueError(
                "Model is None or is not an instance of REFRAMED CBModel")
        self.model = model
        try:
            self.objective = self.model.get_objective() if objective is None else objective
        except Exception:
            self.objective = None
        self.environmental_conditions = OrderedDict() if envcond is None else envcond
        self.constraints = OrderedDict() if constraints is None else constraints
        self.solver = solver
        self._essential_reactions = None
        self._essential_genes = None
        self._reference = reference
        self._gene_to_reaction = None
        self.solver = solver
        self._reset_solver = ModelConstants.RESET_SOLVER
        self.reverse_sintax = [('_b', '_f')]
        self._index_metabolites_reactions = None
        self._m_r_lookup = None

        self.__status_mapping = {
            s_status.OPTIMAL: SStatus.OPTIMAL,
            s_status.UNBOUNDED: SStatus.UNBOUNDED,
            s_status.INFEASIBLE: SStatus.INFEASIBLE,
            s_status.INF_OR_UNB: SStatus.INF_OR_UNB,
            s_status.UNKNOWN: SStatus.UNKNOWN,
            s_status.SUBOPTIMAL: SStatus.SUBOPTIMAL
        }

    @property
    def reference(self):
        """The reference wild type reaction flux values.

        :returns: A dictionary of wild type reaction flux values.

        """
        if self._reference is None:
            self._reference = self.simulate(
                method=SimulationMethod.pFBA).fluxes
        return self._reference

    @property
    def essential_reactions(self, min_growth=0.01):
        """Essential reactions are those when knocked out enable a biomass flux value above a minimal growth defined as
        a percentage of the wild type growth.

        :param float min_growth: Minimal percentage of the wild type growth value. Default 0.01 (1%).
        :returns: A list of essential reactions.

        """
        if self._essential_reactions is not None:
            return self._essential_reactions
        wt_solution = self.simulate()
        wt_growth = wt_solution.objective_value
        reactions = self.model.reactions.keys()
        self._essential_reactions = []
        for rxn in reactions:
            res = self.simulate(constraints={rxn: 0})
            if res:
                if (res.status == SStatus.OPTIMAL and res.objective_value < wt_growth * min_growth) \
                        or res.status == SStatus.INFEASIBLE:
                    self._essential_reactions.append(rxn)
        return self._essential_reactions

    @property
    def essential_genes(self, min_growth=0.01):
        """Essential genes are those when deleted enable a biomass flux value above a minimal growth defined as
        a percentage of the wild type growth.

        :param float min_growth: Minimal percentage of the wild type growth value. Default 0.01 (1%).
        :returns: A list of essential genes.

        """
        if self._essential_genes is not None:
            return self._essential_genes
        self._essential_genes = []
        wt_solution = self.simulate()
        wt_growth = wt_solution.objective_value
        genes = self.model.genes
        for gene in genes:
            active_genes = set(self.model.genes) - set([gene])
            active_reactions = self.evaluate_gprs(active_genes)
            inactive_reactions = set(
                self.model.reactions) - set(active_reactions)
            gr_constraints = {rxn: 0 for rxn in inactive_reactions}
            res = self.simulate(constraints=gr_constraints)
            if res:
                if (res.status == SStatus.OPTIMAL and res.objective_value < wt_growth * min_growth) \
                        or res.status == SStatus.INFEASIBLE:
                    self._essential_genes.append(gene)
        return self._essential_genes

    def evaluate_gprs(self, active_genes):
        """Returns the list of active reactions for a given list of active genes.

        :param list active_genes: List of genes identifiers.
        :returns: A list of active reaction identifiers.

        """
        active_reactions = []
        reactions = self.model.reactions
        for r_id, reaction in reactions.items():
            if reaction.gpr:
                if evaluate_expression_tree(str(reaction.gpr), active_genes):
                    active_reactions.append(r_id)
            else:
                active_reactions.append(r_id)
        return active_reactions

    def update(self):
        self.model.update()

    def add_reaction(self, reaction, replace=True):
        self.model.add_reaction(reaction, replace=replace)

    def remove_reaction(self, r_id):
        self.model.remove_reaction(r_id)

    def get_metabolite_reactions(self, metabolite):
        if not self._index_metabolites_reactions:
            self.__index_metabolites_reactions__()
        return self._index_metabolites_reactions[metabolite]

    def get_metabolite_compartement(self, metabolite):
        return self.model.metabolites[metabolite].compartment

    def get_uptake_reactions(self):
        """
        :returns: The list of uptake reactions.

        """
        drains = self.get_drains()
        reacs = [r for r in drains if self.model.reactions[r].reversible or
                 ((self.model.reactions[r].lb is None or self.model.reactions[r].lb < 0)
                  and len(self.model.reactions[r].get_substrates()) > 0) or
                 ((self.model.reactions[r].ub is None or self.model.reactions[r].ub > 0)
                  and len(self.model.reactions[r].get_products())) > 0]
        return reacs

    def reverse_reaction(self, reaction_id):
        """
        Identify if a reaction is reversible and returns the
        reverse reaction if it is the case.

        :param reaction_id: A reaction identifier.
        :return: A reverse reaction identifier or None

        """

        # TODO: ... use regex instead.

        rxn = self.model.reactions[reaction_id]
        reactions = self.model.reactions
        if rxn.lb < 0:
            rxn.reversible = True
            return reaction_id
        # The model might have been converted to irreversible by REFRAMED in which case reversible reactions
        # are decoupled into forward (reaction_id+'_f') and backward (reaction_id+'_b') reactions
        # or migth be using some other identifier which must be included in self.reverse_sufix
        else:
            for a, b in self.reverse_sintax:
                n = len(reaction_id)-len(a)
                m = len(reaction_id)-len(b)
                if reaction_id[n:] == a and reactions[reaction_id[:n]+b]:
                    return reaction_id[:n]+b
                elif reaction_id[m:] == b and reactions[reaction_id[:m]+a]:
                    return reaction_id[:m]+a
                else:
                    continue
            return None

    def gene_reactions(self):
        """
        :returns: a map of genes to reactions.

        """
        if not self._gene_to_reaction:
            gr = OrderedDict()
            for rxn_id in self.reactions:
                rxn = self.model.reactions[rxn_id]
                if rxn.gpr:
                    genes = rxn.gpr.get_genes()
                    for g in genes:
                        if g in gr.keys():
                            gr[g].append(rxn_id)
                        else:
                            gr[g] = [rxn_id]
            self._gene_to_reaction = gr
        return self._gene_to_reaction

    def get_reactions_for_genes(self, genes):
        """
        Returns the list of reactions catalysed by a list of genes

        :param list genes: A list of gene IDs.
        :returns: A list of reaction identifiers.

        """
        if not self._gene_to_reaction:
            self.gene_reactions()
        reactions = []
        for gene in genes:
            reactions.extend(self._gene_to_reaction[gene])
        return reactions

    def get_reaction_metabolites(self, reaction):
        '''
        Returns all metabolites of a given reaction

        :param reaction: reaction (str)
        :return: metabolites (dict)

        '''
        return self.model.reactions[reaction].stoichiometry

    def is_reactant(self, reaction, metabolite):
        '''
        Returns if a metabolite is reactant into a given reaction

        :param reaction: reaction (str)
        :param metabolite: metabolite (str)
        :return: bool
        '''
        if metabolite not in self.model.reactions[reaction].stoichiometry:
            raise KeyError("{} not in {}".format(metabolite, reaction))
        return self.model.reactions[reaction].stoichiometry[metabolite] < 0.0

    def is_product(self, reaction, metabolite):
        '''
        Returns if a metabolite is product into a given reaction.

        :param reaction: reaction (str)
        :param metabolite: metabolite (str)
        :return: bool
        '''
        if metabolite not in self.model.reactions[reaction].stoichiometry:
            raise KeyError("{} not in {}".format(metabolite, reaction))
        return self.model.reactions[reaction].stoichiometry[metabolite] > 0.0

    def __index_metabolites_reactions__(self):
        self._index_metabolites_reactions = {}
        for reaction in self.reactions:
            metabolites = self.get_reaction_metabolites(reaction)
            for metabolite in metabolites:
                if metabolite in self._index_metabolites_reactions:
                    self._index_metabolites_reactions[metabolite].append(
                        reaction)
                else:
                    self._index_metabolites_reactions[metabolite] = [reaction]

    def metabolite_reaction_lookup(self, force_recalculate=False):
        """ Return the network topology as a nested map from metabolite to reaction to coefficient.
        :return: a dictionary lookup table
        """

        if not self._m_r_lookup or force_recalculate:
            self._m_r_lookup = OrderedDict([(m_id, OrderedDict()) for m_id in self.metabolites])

            for r_id, reaction in self.model.reactions.items():
                for m_id, coeff in reaction.stoichiometry.items():
                    self._m_r_lookup[m_id][r_id] = coeff

        return self._m_r_lookup

    def set_objective(self, reaction):
        self.model.set_objective({reaction: 1})

    def get_objective(self):
        return list(self.objective.keys())

    def get_S(self):
        """
        Returns the S matrix as a numpy array
        :return: S matrix, np.array
        """

        return np.array(self.model.stoichiometric_matrix())

    def get_reaction_bounds(self, reaction):
        """
        Returns the bounds for a given reaction.
        :param reaction: str, reaction ID
        :return: lb(s), ub(s), tuple
        """

        if reaction in self.constraints:
            lb, ub = self.constraints[reaction]
        elif reaction in self.environmental_conditions:
            lb, ub = self.environmental_conditions[reaction]
        else:
            lb, ub = self.model.reactions[reaction].lb, self.model.reactions[reaction].ub

        return lb if lb > -np.inf else -999999, ub if ub < np.inf else 999999

    def get_bounds(self):
        """
        Returns the whole set of lower and upper bounds as numpy arrays

        :return: lb(s), ub(s), tuple of lists

        """
        lbs, ubs = list(zip(*[self.get_reaction_bounds(reaction) for reaction in self.model.reactions]))
        return list(lbs), list(ubs)

    def find_bounds(self):
        """
        Return the median upper and lower bound of the metabolic model.
        Bounds can vary from model to model. Cobrapy defaults to (-1000, 1000).
        """
        lower_bounds = np.asarray([rxn.lb for rxn in self.model.reactions], dtype=float)
        upper_bounds = np.asarray([rxn.ub for rxn in self.model.reactions], dtype=float)
        lower_bound = np.nanmedian(lower_bounds[lower_bounds != 0.0])
        upper_bound = np.nanmedian(upper_bounds[upper_bounds != 0.0])
        if np.isnan(lower_bound):
            LOGGER.warning("Could not identify a median lower bound.")
            lower_bound = -1000.0
        if np.isnan(upper_bound):
            LOGGER.warning("Could not identify a median upper bound.")
            upper_bound = 1000.0
        return lower_bound, upper_bound
        
    def find_unconstrained_reactions(self):
        """Return list of reactions that are not constrained at all."""
        lower_bound, upper_bound = self.find_bounds()
        return [
            rxn
            for rxn in self.model.reactions
            if rxn.lb <= lower_bound and rxn.ub >= upper_bound
        ]


    def get_boundary_reaction(self, metabolite):

        """
        Finds the boundary reaction associated with an extracellular metabolite.
        If none is found, None is returned

        :param metabolite: str, metabolite ID

        :returns: reaction, str or  None
        """

        # reaction.reaction_type.value == 'exchange'
        # len(reaction.stoichiometry) == 1

        for reaction in self.get_metabolite_reactions(metabolite):

            if reaction in self.medium or self.model.reactions[reaction].reaction_type.value == 'exchange' \
                    or len(self.model.reactions[reaction].stoichiometry) == 1:
                return reaction

        return None

    # Simulate
    def simulate(self, objective=None, method=SimulationMethod.FBA,
                 maximize=True, constraints=None, reference=None,
                 scalefactor=None, solver=None):
        '''
        Simulates a phenotype when applying a set constraints using the specified method.

        :param dic objective: The simulation objective. If none, the model objective is considered.
        :param method: The SimulationMethod (FBA, pFBA, lMOMA, etc ...)
        :param boolean maximize: The optimization direction
        :param dic constraints: A dictionary of constraints to be applied to the model.
        :param dic reference: A dictionary of reaction flux values.
        :param float scalefactor: A positive scaling factor for the solver. Default None.
        :param solver: An instance of the solver.
        '''

        a_solver = solver

        if not objective:
            objective = self.objective

        simul_constraints = OrderedDict()
        if constraints:
            simul_constraints.update(constraints)
        if self.constraints:
            simul_constraints.update(self.constraints)
        if self.environmental_conditions:
            simul_constraints.update(self.environmental_conditions)

        if not a_solver and not self._reset_solver:
            if self.solver is None:
                self.solver = solver_instance(self.model)
            a_solver = self.solver

        # scales the model if a scalling factor is defined.
        # ... for now resets the solver...
        # ... scalling should be implemented at the solver level.
        if scalefactor:
            a_solver = None
            for _, rxn in self.model.reactions.items():
                rxn.lb = rxn.lb * scalefactor
                rxn.ub = rxn.ub * scalefactor
            if simul_constraints:
                for idx, constraint in simul_constraints.items():
                    if isinstance(constraint, (int, float)):
                        simul_constraints[idx] = constraint * scalefactor
                    elif isinstance(constraint, tuple):
                        simul_constraints[idx] = tuple(
                            x * scalefactor for x in constraint)
                    else:
                        raise ValueError("Could not scale the model")

        # TODO: simplifly ...
        if method in [SimulationMethod.lMOMA, SimulationMethod.MOMA, SimulationMethod.ROOM] and reference is None:
            reference = self.reference

        if method == SimulationMethod.FBA:
            solution = FBA(self.model,  objective=objective, minimize=not maximize,
                           constraints=simul_constraints, solver=a_solver)
        elif method == SimulationMethod.pFBA:
            solution = pFBA(self.model, objective=objective, minimize=not maximize,
                            constraints=simul_constraints, solver=a_solver, obj_frac=0.999)
        elif method == SimulationMethod.lMOMA:
            solution = lMOMA(self.model, constraints=simul_constraints,
                             reference=reference, solver=a_solver)
        elif method == SimulationMethod.MOMA:
            solution = MOMA(self.model,  constraints=simul_constraints,
                            reference=reference, solver=a_solver)
        elif method == SimulationMethod.ROOM:
            solution = ROOM(self.model,  constraints=simul_constraints,
                            reference=reference, solver=a_solver)
        # Special case in which only the simulation context is required without any simulatin result
        elif method == SimulationMethod.NONE:
            solution = Solution(status=s_status.UNKNOWN,
                                message=None, fobj=None, values=None)
        else:
            raise Exception(
                "Unknown method to perform the simulation.")

        # undoes the model scaling
        if scalefactor:
            for _, rxn in self.model.reactions.items():
                rxn.lb = rxn.lb / scalefactor
                rxn.ub = rxn.ub / scalefactor
            if solution.status in (s_status.OPTIMAL, s_status.SUBOPTIMAL):
                solution.fobj = solution.fobj / scalefactor
                for x, y in solution.values.items():
                    solution.values[x] = y / scalefactor

        status = self.__status_mapping[solution.status]

        result = SimulationResult(self.model, solution.fobj, fluxes=solution.values, status=status,
                                  envcond=self.environmental_conditions, model_constraints=self.constraints,
                                  simul_constraints=constraints, maximize=maximize)
        return result

    def FVA(self, obj_frac=0.9, reactions=None, constraints=None, loopless=False, internal=None, solver=None):
        """ Flux Variability Analysis (FVA).

        :param model: An instance of a constraint-based model.
        :param float obj_frac: The minimum fraction of the maximum growth rate (default 0.9). Requires that the objective value is at least the fraction times maximum objective value. A value of 0.85 for instance means that the objective has to be at least at 85% percent of its maximum.
        :param list reactions: List of reactions to analyze (default: all).
        :param dic constraints: Additional constraints (optional).
        :param boolean loopless: Run looplessFBA internally (very slow) (default: false).
        :param list internal: List of internal reactions for looplessFBA (optional).
        :param solver: A pre-instantiated solver instance (optional)
        :returns: A dictionary of flux variation ranges.

        """
        _constraints = {}
        if constraints:
            _constraints.update(constraints)
        if self.environmental_conditions:
            _constraints.update(self.environmental_conditions)
        from reframed.cobra.variability import FVA
        return FVA(self.model, obj_frac=obj_frac, reactions=reactions,
                   constraints=_constraints, loopless=loopless, internal=internal, solver=solver)


class GeckoSimulation(Simulation):

    def __init__(self, model: GeckoModel, objective=None, envcond=None, constraints=None,  solver=None, reference=None):
        super(GeckoSimulation, self).__init__(
            model, objective, envcond, constraints, solver, reference)
        self._essential_proteins = None

    @property
    def proteins(self):
        return self.model.proteins

    @property
    def protein_rev_reactions(self):
        return self.model.protein_rev_reactions

    def essential_proteins(self, protein_prefix, min_growth=0.01):
        if self._essential_proteins is not None:
            return self._essential_proteins
        wt_solution = self.simulate()
        wt_growth = wt_solution.objective_value
        self._essential_proteins = []
        proteins = self.model.proteins
        for p in proteins:
            rxn = "{}{}".format(protein_prefix, p)
            res = self.simulate(constraints={rxn: 0})
            if res:
                if (res.status == SStatus.OPTIMAL and res.objective_value < wt_growth * min_growth) \
                        or res.status == SStatus.INFEASIBLE:
                    self._essential_proteins.append(rxn)
        return self._essential_proteins

    def protein_reactions(self, protein):
        """
        Returns the list of reactions associated to a protein
        """
        reactions = []
        for r_id, rxn in self.model.reactions.items():
            lsub = rxn.get_substrates()
            for m in lsub:
                if protein in m:
                    reactions.append(r_id)
        return reactions

    def reverse_reaction(self, reaction_id):
        """
        Identify if a reaction is reversible and returns the
        reverse reaction if it is the case.

        :returns: A reaction identifier or None.
        """
        f, d = zip(*self.model.protein_rev_reactions.values())
        if reaction_id in f:
            return d[f.index(reaction_id)]
        elif reaction_id in d:
            return f[d.index(reaction_id)]
        else:
            return None




from mewpy.regulation import SRFBAModel
import time
import warnings
import os

warnings.filterwarnings("ignore")


def sample_network():
    """
    Sample network available in examples using cobra as reader

    :return: sample sr-fba model
    """

    # Sample network

    import cobra.io
    from mewpy.simulation.cobra import Simulation

    DIR = os.path.dirname(os.path.realpath(__file__)) 
    cbm_model_f = os.path.join(DIR,'../../../examples/models/regulation/SampleNet.xml')
    reg_model_f = os.path.join(DIR,'../../../examples/models/regulation/SampleRegNet.csv')

    _BIOMASS_ID = 'r11'

    t0 = time.time()
    model = cobra.io.read_sbml_model(cbm_model_f)
    model.objective = _BIOMASS_ID

    simulation = Simulation(model)
    t1 = time.time()

    print(t1 - t0)

    t0 = time.time()
    srfba = SRFBAModel.from_tabular_format(reg_model_f, model, cbm_simulation_interface=simulation,
                                           sep=',', id_col=0, rule_col=1, header=None)
    t1 = time.time()

    print(t1 - t0)

    # r15
    initial_state = {
        'pH': 7,
        'g21': 1,
        'g22': 0,
        'g23': 1,
        'g24': 1,
        'g25': 1,
        'g26': 1,
        'g27': 1,
        'g28': 1,
        'g29': 1,
        'g30': 0,
        'g31': 0,
        'g32': 0,
        'g33': 1,
        'g36': 1,
    }

    # # not r15
    # initial_state = {
    #     'pH': 7,
    #     'g21': 1,
    #     'g22': 0,
    #     'g23': 1,
    #     'g24': 0,
    #     'g25': 0,
    #     'g26': 0,
    #     'g27': 1,
    #     'g28': 1,
    #     'g29': 1,
    #     'g30': 0,
    #     'g31': 1,
    #     'g32': 0,
    #     'g33': 1,
    #     'g36': 0,
    # }

    srfba.initial_state = initial_state

    t0 = time.time()
    srfba.simulate()
    t1 = time.time()

    print(t1 - t0)

    return srfba


def cobra_ecoli_core_model():
    """

    SRFBA COBRA Ecoli Core Model (subset of the genome-scale metabolic reconstruction iAF1260)

    Available at Orth J, Fleming R, Palsson B. 2010. Reconstruction and Use of Microbial Metabolic Networks:
    the Core Escherichia coli Metabolic Model as an Educational Guide, EcoSal Plus 2010;
    doi:10.1128/ecosalplus.10.2.1

    Ecoli core model from cobra.test
    Regulatory model from http://systemsbiology.ucsd.edu/Downloads/EcoliCore

    :return: srfba model
    """

    import cobra.test
    from mewpy.simulation.cobra import Simulation

    DIR = os.path.dirname(os.path.realpath(__file__))
    reg_model_f = os.path.join(DIR,'../../../examples/models/regulation/core_TRN_v2.csv')
    aliases_f = os.path.join(DIR,'../../../examples/models/regulation/core_TRN_srfba_aliases.csv')

    _BIOMASS_ID = 'Biomass_Ecoli_core'
    _O2 = 'EX_o2_e'
    _GLC = 'EX_glc__D_e'
    _FUM = 'EX_fum_e'

    t0 = time.time()
    model = cobra.test.create_test_model("textbook")
    model.objective = _BIOMASS_ID

    envcond = {_GLC: (-10.0, 100000.0), _O2: (-30, 100000.0), _FUM: (-10, 100000.0)}
    simulation = Simulation(model, envcond=envcond)
    t1 = time.time()

    print(t1 - t0)

    t0 = time.time()
    srfba = SRFBAModel.from_tabular_format(reg_model_f, model, cbm_simulation_interface=simulation,
                                           sep=',', id_col=1, rule_col=2, aliases_cols=[0], header=0)
    srfba.update_aliases_from_tabular_format_file(aliases_f, id_col=1, aliases_cols=[0])
    t1 = time.time()

    print(t1 - t0)

    initial_state = {
        'Biomass_Ecoli_core_w_GAM': 1
    }

    srfba.initial_state = initial_state

    t0 = time.time()
    srfba.simulate()
    t1 = time.time()

    print(t1 - t0)

    return srfba


def framed_imc1010_model():
    """

    SRFBA FRAMED Ecoli iMC1010 model from Covert et al. 2004 at https://doi.org/10.1038/nature02456
    This model uses the E. coli metabolic model iJR904 available at https://www.ebi.ac.uk/biomodels/MODEL1507180060
    and published at https://doi.org/10.1186/gb-2003-4-9-r54

    Some rules had to be adjusted though
    iJR904 had to be adjusted, as it didn't match SR_FBA original publication or had errors

    The following reactions were added as in the original publication:
        - h2so: 2 o2_c + h2s_c -> (0, 999999) so4_c + 2 h_c
        - h2st: h2s_e <-> (-999999, 999999) h2s_c
        - h2s_ext: h2s_e -> (0, 999999)
        - 5dglcn_ext: 5dglcn_e -> (0, 999999)
        - btn_ext: btn_e -> (0, 999999)
        - cbi_ext: cbi_e -> (0, 999999)
        - h2o2_ext: h2o2_e -> (0, 999999)
        - ppa_ext: ppa_e -> (0, 999999)
        - thym_ext: thym_e-> (0, 999999)

    The following reactions were removed from iJR904, as they were duplicated or wrong
        - GALU (duplicated GALUi (correct GPR))

    The following reactions were removed from SR-FBA orginal publication model, as they were duplicated or wrong:
        - ptrca (duplicated PTRCTA)
        - indolet (unbalanced and duplicated indolet2r)

    The following indirect mappings were found:

        cbiat, R_CBIAT_DELETE
        cblat, R_CBLAT_DELETE
        d-lact2, R_D_LACt2
        dhptdc, R_DHPTDCs
        l-lacd2, R_L_LACD2
        l-lact2r, R_L_LACt2r
        lcar, R_LCARS
        nh3t, R_NH4t
        l-lacd3, R_L_LACD3
        test_akgdh, R_AKGDH
        test_nadtrhd, R_NADTRHD
        trpas1, CYSDS

    Others:
        - iJR904 has an additional reaction to ptrcta, namely ORNTA


    :return: srfba model
    """

    from reframed.io.sbml import load_cbmodel
    from mewpy.simulation.reframed import Simulation

    DIR = os.path.dirname(os.path.realpath(__file__))
    cbm_model_f = os.path.join(DIR,"../../../examples/models/iJR904_srfba.xml")
    reg_model_f = os.path.join(DIR,'../../../examples/models/regulation/imc1010_v6.csv')
    aliases_f = os.path.join(DIR,'../../../examples/models/regulation/imc1010_srfba_aliases.csv')

    _BIOMASS_ID = 'R_BiomassEcoli'
    _O2 = 'R_EX_o2_e'
    _GLC = 'R_EX_glc_DASH_D_e'
    _CO2 = 'R_EX_co2_e'
    _FE2 = "R_EX_fe2_e"
    _H = "R_EX_h_e"
    _H2O = "R_EX_h2o_e"
    _K = "R_EX_k_e"
    _NA1 = "R_EX_na1_e"
    _NH4 = "R_EX_nh4_e"
    _PI = "R_EX_pi_e"
    _SO4 = "R_EX_so4_e"
    _SUCC = "R_EX_succ_e"

    t0 = time.time()
    model = load_cbmodel(cbm_model_f)
    model.set_objective({_BIOMASS_ID: 1})

    # envcond = {_GLC: (-10.0, 100000.0),
    #            _O2: (-100.0, 100000.0),
    #            _CO2: (-100.0, 100000.0),
    #            _FE2: (-100.0, 100000.0),
    #            _H: (-100.0, 100000.0),
    #            _H2O: (-100.0, 100000.0),
    #            _K: (-100.0, 100000.0),
    #            _NA1: (-100.0, 100000.0),
    #            _NH4: (-100.0, 100000.0),
    #            _PI: (-100.0, 100000.0),
    #            _SO4: (-100.0, 100000.0)}

    envcond = {_GLC: (-10.0, 100000.0),
               _SUCC: (0.0, 100000.0),
               _NH4: (-10.0, 100000.0),
               _O2: (-10.0, 100000.0),
               _CO2: (-15.0, 100000.0),
               _PI: (-15.0, 100000.0),
               _SO4: (-10.0, 100000.0),
               _H: (-10.0, 100000.0),
               _H2O: (-55.0, 100000.0)}

    simulation = Simulation(model, envcond=envcond)
    t1 = time.time()

    print(t1 - t0)

    t0 = time.time()
    srfba = SRFBAModel.from_tabular_format(reg_model_f, model, cbm_simulation_interface=simulation,
                                           sep=';', id_col=1, rule_col=4, aliases_cols=[0, 2, 3], header=0)
    srfba.update_aliases_from_tabular_format_file(aliases_f, sep=';', id_col=0, aliases_cols=[1])
    t1 = time.time()

    print(t1 - t0)

    m_initial_state = {i.id: 1 for i in srfba.metabolic_regulatory_genes_gen()}

    initial_state = {
        'Growth': 1,
        'pH': 7
    }

    initial_state.update(m_initial_state)

    srfba.initial_state = initial_state

    t0 = time.time()
    srfba.simulate()
    t1 = time.time()

    print(t1 - t0)

    return srfba


if __name__ == '__main__':
    srfba_sample_net = sample_network()
    srfba_ecoli_core = cobra_ecoli_core_model()
    srfba_ecoli_imc1010 = framed_imc1010_model()



from mewpy.regulation.optram import OptRAMRegModel, OptRamProblem, load_optram
from mewpy.simulation import SimulationMethod, get_simulator
from mewpy.optimization.evaluation import BPCY_FVA, BPCY, WYIELD
from mewpy.optimization import EA
from time import time
import os


def test():
    dir_path = os.path.dirname(os.path.realpath(__file__))
    PATH = os.path.join(dir_path, '../../../examples/models/optram/')
    gene_file = os.path.join(PATH, 'mgene.csv')
    ft_file = os.path.join(PATH, 'TFnames.csv')
    matrix_file = os.path.join(PATH, 'regnet.csv')
    model_file = os.path.join(PATH, 'yeast_7.6-optram.xml')

    BIOMASS_ID = 'r_2111'
    PRODUCT_ID = 'r_1912'
    GLC = 'r_1714'

    envcond = {GLC:(-10,0)}

    # adds the prefix 'G_' to genes. Only for REFRAMED models
    regnet = load_optram(gene_file, ft_file, matrix_file, gene_prefix='G_')
    #from cobra.io import read_sbml_model
    #model = read_sbml_model(model_file)
    
    # the general objective is to maximize the target
    from reframed.io.sbml import load_cbmodel
    model = load_cbmodel(model_file)
    model.set_objective({BIOMASS_ID:1})
    
    evaluator_1 = BPCY(BIOMASS_ID, PRODUCT_ID, method=SimulationMethod.lMOMA)
    evaluator_2 = WYIELD(BIOMASS_ID, PRODUCT_ID, parsimonious=True)
    
    # OptRAM problem
    problem = OptRamProblem(model, [evaluator_1, evaluator_2],
                            regnet, envcond = envcond, candidate_min_size=10, candidate_max_size=30)

    #print(problem.target_list)
    #print("\n\n")
    #print(problem.simulator.genes)    
    #print("\n\n")
    #print(set(problem.target_list).intersection(set(problem.simulator.genes)))


    ea = EA(problem, max_generations=3, mp=True)
    final_pop = ea.run()
    import mewpy.utils.utilities as utl
    millis = int(round(time() * 1000))
    filename = "OPTRAM{}_OU_{}.csv".format('TRP', millis)
    utl.population_to_csv(problem, final_pop, filename, simplify=False)


if __name__ == "__main__":
    for _ in range(1):
        test()



from warnings import warn
from mewpy.io import read_tabular_aliases
from mewpy.regulation.regulatory_interaction import RegulatoryInteraction
from mewpy.regulation.regulatory_variable import RegulatoryVariable
from sympy.logic.boolalg import BooleanTrue, BooleanFalse
from sympy.core.numbers import Zero, One


class RegulatoryModel(object):

    def __init__(self, identifier,
                 name=None,
                 targets=None,
                 regulators=None,
                 regulatory_interactions=None):

        if not identifier:
            raise BaseException("An ID must be provided")

        if not name:
            name = identifier

        if not targets:
            targets = {}

        if not regulators:
            regulators = {}

        if not regulatory_interactions:
            regulatory_interactions = {}

        self._id = None

        self._name = None

        self._targets = None

        self._regulators = None

        # TODO: add regulator and target should change the property state
        self._regulatory_variables = {}

        self._regulatory_interactions = None

        self._regulatory_solution = []

        self.__build_regulatory_model__(identifier=identifier,
                                        name=name,
                                        targets=targets,
                                        regulators=regulators,
                                        regulatory_interactions=regulatory_interactions)

    def __str__(self):
        return self._name

    def __repr__(self):

        return '{}({}, {}) at {}'.format(self.__class__.__name__, self._id, self._name, id(self))

    def __build_regulatory_model__(self, identifier,
                                   name,
                                   targets,
                                   regulators,
                                   regulatory_interactions):

        if targets:

            if not isinstance(targets, dict):
                raise TypeError(
                    "Targets must be a dict of RegulatoryVariables objects")

            if not isinstance(targets[list(targets.keys())[0]], RegulatoryVariable):
                raise TypeError(
                    "Targets must be a dict of RegulatoryVariables objects")

            self._targets = targets

        else:
            self._targets = {}

        if regulators:

            if not isinstance(regulators, dict):
                raise TypeError("Regulators must be a dict of RegulatoryVariables objects")

            if not isinstance(regulators[list(regulators.keys())[0]], RegulatoryVariable):
                raise TypeError("Regulators must be a dict of RegulatoryVariables objects")

            self._regulators = regulators

        else:
            self._regulators = {}

        if regulatory_interactions:

            if not isinstance(regulatory_interactions, dict):
                raise TypeError(
                    "Regulatory interactions must be a dict of RegulatoryInteraction objects")

            if not isinstance(regulatory_interactions[list(regulatory_interactions.keys())[0]], RegulatoryInteraction):
                raise TypeError("Regulatory interactions must be a dict of RegulatoryInteraction objects. "
                                "Use RegulatoryInteraction() to convert")

            self._regulatory_interactions = regulatory_interactions

        else:
            self._regulatory_interactions = {}

        self.id = identifier
        self.name = name

    @classmethod
    def from_dict(cls, identifier,
                  name=None,
                  targets=None,
                  regulators=None,
                  regulatory_interactions=None):
        # simulation_interface=None):

        return cls(identifier, name, targets, regulators, regulatory_interactions)  # , simulation_interface)

    @property
    def id(self):
        return getattr(self, '_id', None)

    @property
    def name(self):
        return getattr(self, '_name', None)

    @property
    def targets(self):
        return getattr(self, '_targets', None)

    @property
    def regulators(self):
        return getattr(self, '_regulators', None)

    @property
    def regulatory_variables(self):
        return getattr(self, '_regulatory_variables', None)

    @property
    def regulatory_interactions(self):
        return getattr(self, '_regulatory_interactions', None)

    @property
    def regulatory_solution(self):
        return getattr(self, '_regulatory_solution', None)

    @id.setter
    def id(self, value):

        if value == self._id:
            pass

        elif not isinstance(value, str):
            raise TypeError("The ID must be a string. Use str() to convert")

        else:
            self._id = value

    @name.setter
    def name(self, value):

        if value == self._name:
            pass

        elif not isinstance(value, str):
            raise TypeError("The name must be a string. Use str() to convert")

        else:
            self._name = value

    @targets.setter
    def targets(self, value):

        if value == self._targets:
            pass

        else:
            raise BaseException("Targets property cannot be changed. "
                                "To add new targets to the model, please use the add_new_target method.")

    @regulators.setter
    def regulators(self, value):

        if value == self._regulators:
            pass

        else:
            raise BaseException("Regulators property cannot be changed. "
                                "To add new regulators to the model, please use the add_new_regulator method.")

    @regulatory_variables.setter
    def regulatory_variables(self, value):

        if value == self._regulatory_variables:
            pass

        else:
            raise BaseException("regulatory variables property cannot be changed. "
                                "To add new targets or regulators to the model, please use the add_new_target method.")

    @regulatory_interactions.setter
    def regulatory_interactions(self, value):

        if value == self._regulatory_interactions:
            pass

        elif not value:
            self._regulatory_interactions = {}

        else:

            if not isinstance(value, dict):
                raise TypeError(
                    "Regulatory interactions must be a dict of RegulatoryInteraction objects")

            if not isinstance(value[list(value.keys())[0]], RegulatoryInteraction):
                raise TypeError("Regulatory interactions must be a dict of RegulatoryInteraction objects. "
                                "Use RegulatoryInteraction() to convert")

            def regulatory_interactions(message):
                warn(message, UserWarning, stacklevel=2)

            regulatory_interactions("Setting/altering the regulatory interactions associated with a model replaces the "
                                    "whole model. Regulators and targets will be ignored and rebuild. "
                                    "To add new regulatory interactions to the stack use the "
                                    "add_new_regulatory_interaction method")

            self._regulatory_interactions = value

            for reg_interaction in value.values():

                for reg_val in reg_interaction.regulators_gen():

                    if reg_val.id not in self._regulators:
                        self._regulators[reg_val.id] = reg_val

                if reg_interaction.target.id not in self._targets:
                    self._targets[reg_interaction.target.id] = reg_interaction.target

    def get_regulator(self, identifier):

        if not isinstance(identifier, str):
            raise TypeError("The ID must be a string. Use str() to convert")

        return self._regulators[identifier]

    def get_target(self, identifier):

        if not isinstance(identifier, str):
            raise TypeError("The ID must be a string. Use str() to convert")

        return self._targets[identifier]

    def get_regulatory_variable(self, identifier):

        if not isinstance(identifier, str):
            raise TypeError("The ID must be a string. Use str() to convert")

        if identifier in self.regulators:
            return self._regulators[identifier]
        elif identifier in self.targets:
            return self._targets[identifier]
        else:
            return None

    def get_regulatory_interaction(self, identifier):

        if not isinstance(identifier, str):
            raise TypeError("The ID must be a string. Use str() to convert")

        return self._regulatory_interactions[identifier]

    def regulators_gen(self):

        return (value for value in self._regulators.values())

    def targets_gen(self):

        return (value for value in self._targets.values())

    def regulatory_variables_gen(self):

        return (value for value in self._regulatory_variables.values())

    def regulatory_interactions_gen(self):

        return (value for value in self._regulatory_interactions.values())

    def simulate(self):

        raise NotImplementedError("This is the base classe for type-specific regulatory models. Instantiate these "
                                  "instead to perform simulations")

    def add_regulator(self, identifier,
                      name=None,
                      expression_coef=None,
                      aliases=None,
                      regulatory_interactions=None,
                      targets=None,
                      regulators=None):

        if not isinstance(identifier, str):
            raise TypeError("The ID must be a string. Use str() to convert")

        if identifier in self.regulators:
            def add_regulator_warn(message):
                warn(message, UserWarning, stacklevel=2)

            add_regulator_warn("The ID already exists in the dictionary of regulators of the regulatory "
                               "model. Ignoring the alteration")

            return self.regulators[identifier]

        if identifier in self.targets:
            self.targets[identifier].is_regulator = True

            return self.targets[identifier]

        regulator = RegulatoryVariable(
            identifier=identifier,
            name=name,
            expression_coef=expression_coef,
            aliases=aliases,
            regulatory_interactions=regulatory_interactions,
            targets=targets,
            regulators=regulators,
            model=self,
            is_regulator=True,
            is_target=False)

        self.regulatory_variables[identifier] = regulator

        return regulator

    def add_regulator_object(self, regulator):

        if not isinstance(regulator, RegulatoryVariable):
            raise TypeError("Regulator must be RegulatoryVariable object")

        if regulator.id in self.regulators:
            def add_regulator_warn(message):
                warn(message, UserWarning, stacklevel=2)

            add_regulator_warn("The ID already exists in the dictionary of regulators of the regulatory "
                               "model. Ignoring the alteration")
            return

        if regulator.id in self.targets:
            regulator.is_target = True

        for reg_inter in regulator.regulatory_interactions_gen():
            reg_inter.model = self

        for reg in regulator.regulators_gen():
            reg.model = self

        for tar in regulator.targets_gen():
            tar.model = self

        regulator.model = self

    def __add_regulator__(self, identifier,
                          name=None,
                          expression_coef=None,
                          aliases=None,
                          regulatory_interactions=None,
                          targets=None,
                          regulators=None):

        if not isinstance(identifier, str):
            raise TypeError("The ID must be a string. Use str() to convert")

        if identifier in self.regulators:

            regulator = self.regulators[identifier]

            if name is not None:
                regulator.name = name
            if expression_coef is not None:
                regulator.expression_coef = expression_coef
            if aliases is not None:
                for alias in aliases:
                    if alias not in regulator.aliases:
                        regulator.aliases.append(alias)
            if regulatory_interactions is not None:
                regulator.regulatory_interactions.update(regulatory_interactions)
            if targets is not None:
                regulator.outputs.update(targets)
            if regulators is not None:
                regulator.inputs.update(regulators)

            return regulator

        if identifier in self.targets:

            regulator = self.targets[identifier]

            if name is not None:
                regulator.name = name
            if expression_coef is not None:
                regulator.expression_coef = expression_coef
            if aliases is not None:
                for alias in aliases:
                    if alias not in regulator.aliases:
                        regulator.aliases.append(alias)
            if regulatory_interactions is not None:
                regulator.regulatory_interactions.update(regulatory_interactions)
            if targets is not None:
                regulator.outputs.update(targets)
            if regulators is not None:
                regulator.inputs.update(regulators)
            regulator.is_regulator = True

            return regulator

        regulator = RegulatoryVariable(
            identifier=identifier,
            name=name,
            expression_coef=expression_coef,
            aliases=aliases,
            regulatory_interactions=regulatory_interactions,
            targets=targets,
            regulators=regulators,
            model=self,
            is_regulator=True,
            is_target=False)

        self.regulatory_variables[identifier] = regulator

        return regulator

    def remove_regulator(self, identifier):

        if identifier not in self.regulators:
            def remove_regulator_id(message):
                warn(message, UserWarning, stacklevel=2)

            remove_regulator_id("The ID does not exist in the dictionary of regulators of the regulatory "
                                "model. Ignoring the alteration")

            return

        for key in self.regulators[identifier].regulatory_interactions.keys():
            self.remove_regulatory_interaction(key)

        if identifier in self.regulators:
            del self.regulators[identifier]

        if identifier in self.targets:
            del self.targets[identifier]

    def add_target(self, identifier,
                   name=None,
                   expression_coef=None,
                   aliases=None,
                   regulatory_interactions=None,
                   targets=None,
                   regulators=None):

        if not isinstance(identifier, str):
            raise TypeError("The ID must be a string. Use str() to convert")

        if identifier in self.targets:
            def add_target_warn(message):
                warn(message, UserWarning, stacklevel=2)

            add_target_warn("The ID already exists in the dictionary of targets of the regulatory "
                            "model. Ignoring the alteration")

            return self.targets[identifier]

        if identifier in self.regulators:
            self.regulators[identifier].is_target = True

            return self.regulators[identifier]

        target = RegulatoryVariable(
            identifier=identifier,
            name=name,
            expression_coef=expression_coef,
            aliases=aliases,
            regulatory_interactions=regulatory_interactions,
            targets=targets,
            regulators=regulators,
            model=self,
            is_regulator=False,
            is_target=True)

        self.regulatory_variables[identifier] = target

        return target

    def add_target_object(self, target):

        if not isinstance(target, RegulatoryVariable):
            raise TypeError("Target must be RegulatoryVariable object")

        if target.id in self.targets:
            def add_regulator_warn(message):
                warn(message, UserWarning, stacklevel=2)

            add_regulator_warn("The ID already exists in the dictionary of targets of the regulatory "
                               "model. Ignoring the alteration")
            return

        if target.id in self.regulators:
            target.is_regulator = True

        for reg_inter in target.regulatory_interactions_gen():
            reg_inter.model = self

        for reg in target.regulators_gen():
            reg.model = self

        for tar in target.targets_gen():
            tar.model = self

        target.model = self

    def __add_target__(self, identifier,
                       name=None,
                       expression_coef=None,
                       aliases=None,
                       regulatory_interactions=None,
                       targets=None,
                       regulators=None):

        if not isinstance(identifier, str):
            raise TypeError("The ID must be a string. Use str() to convert")

        if identifier in self.targets:

            target = self.targets[identifier]

            if name is not None:
                target.name = name
            if expression_coef is not None:
                target.expression_coef = expression_coef
            if aliases is not None:
                for alias in aliases:
                    if alias not in target.aliases:
                        target.aliases.append(alias)
            if regulatory_interactions is not None:
                target.regulatory_interactions.update(regulatory_interactions)
            if targets is not None:
                target.outputs.update(targets)
            if regulators is not None:
                target.inputs.update(regulators)

            return target

        if identifier in self.regulators:

            target = self.regulators[identifier]

            if name is not None:
                target.name = name
            if expression_coef is not None:
                target.expression_coef = expression_coef
            if aliases is not None:
                for alias in aliases:
                    if alias not in target.aliases:
                        target.aliases.append(alias)
            if regulatory_interactions is not None:
                target.regulatory_interactions.update(regulatory_interactions)
            if targets is not None:
                target.outputs.update(targets)
            if regulators is not None:
                target.inputs.update(regulators)

            target.is_target = True

            return target

        target = RegulatoryVariable(
            identifier=identifier,
            name=name,
            expression_coef=expression_coef,
            aliases=aliases,
            regulatory_interactions=regulatory_interactions,
            targets=targets,
            regulators=regulators,
            model=self,
            is_regulator=False,
            is_target=True)

        self.regulatory_variables[identifier] = target

        return target

    def remove_target(self, identifier):

        if identifier not in self.targets:
            def remove_target_id(message):
                warn(message, UserWarning, stacklevel=2)

            remove_target_id("The ID does not exist in the dictionary of targets of the regulatory "
                             "model. Ignoring the alteration")

            return

        for key in self.targets[identifier].regulatory_interactions.keys():
            self.remove_regulatory_interaction(key)

        if identifier in self.targets:
            del self.targets[identifier]

        if identifier in self.regulators:
            del self.regulators[identifier]

    def add_regulatory_interaction(self, identifier,
                                   rule,
                                   target,
                                   target_aliases=None,
                                   regulators=None):

        if not isinstance(identifier, str):
            raise TypeError("The ID must be a string. Use str() to convert")

        if not isinstance(rule, str):
            raise TypeError(
                "The regulatory rule must be a string. Use str() to convert")

        if not target_aliases:
            target_aliases = []

        if not isinstance(target, str) and not isinstance(target, RegulatoryVariable):
            raise TypeError(
                "Target must be a string or a RegulatoryVariable object")

        if identifier in self.regulatory_interactions:
            def add_reg_inter_warn(message):
                warn(message, UserWarning, stacklevel=2)

            add_reg_inter_warn("The ID already exists in the dictionary of regulatory interactions of the regulatory "
                               "model. Ignoring the alteration")
            return

        return RegulatoryInteraction.from_str(identifier=identifier,
                                              rule=rule,
                                              target=target,
                                              target_aliases=target_aliases,
                                              regulators=regulators,
                                              model=self)

    def add_regulatory_interaction_object(self, regulatory_interaction):

        if not isinstance(regulatory_interaction, RegulatoryInteraction):
            raise TypeError(
                "The regulatory interaction must be RegulatoryInteraction object")

        if regulatory_interaction.id in self.regulatory_interactions:
            def add_reg_inter_warn(message):
                warn(message, UserWarning, stacklevel=2)

            add_reg_inter_warn("The ID already exists in the dictionary of regulatory interactions of the regulatory "
                               "model. Ignoring the alteration")
            return

        for reg in regulatory_interaction.regulators_gen():
            reg.model = self

        regulatory_interaction.target.model = self

        regulatory_interaction.model = self

    def remove_regulatory_interaction(self, identifier):

        # TODO: Remove regulator, target and regulatoy_interaction don't follow a destructive approach.
        #  Even though their direct elements (regulatory interactions, targets and regulators) are removed,
        #  the regulatory interactions of these direct elements are nonetheless kept in the regulatory model.
        #  Implement a destructive flag, that destroys any relation with a regulatoy interaction, regulator and target.

        if identifier not in self.regulatory_interactions:
            def remove_reg_inter_id(message):
                warn(message, UserWarning, stacklevel=2)

            remove_reg_inter_id("The ID does not exist in the dictionary of regulatory interactions of the regulatory "
                                "model. Ignoring the alteration")

            return

        target = self.regulatory_interactions[identifier].target

        for tar in target.targets_gen():
            if tar.id in self.targets:
                del self.targets[tar.id]
            if tar.id in self.regulators:
                del self.regulators[tar.id]

        for reg in target.regulators_gen():
            if reg.id in self.regulators:
                del self.regulators[reg.id]
            if reg.id in self.targets:
                del self.targets[reg.id]

        for regulator in self.regulatory_interactions[identifier].regulators_gen():

            for tar in regulator.targets_gen():
                if tar.id in self.targets:
                    del self.targets[tar.id]
                if tar.id in self.regulators:
                    del self.regulators[tar.id]

            for reg in regulator.regulators_gen():
                if reg.id in self.regulators:
                    del self.regulators[reg.id]
                if reg.id in self.targets:
                    del self.targets[reg.id]

        if identifier in self.regulatory_interactions:
            del self.regulatory_interactions[identifier]

    def remove_regulatory_variable(self, identifier):

        if identifier not in self.targets and identifier not in self.regulators:
            def remove_reg_var_id(message):
                warn(message, UserWarning, stacklevel=2)

            remove_reg_var_id("The ID does not exist in the dictionary of targets and regulators of the regulatory "
                              "model. Ignoring the alteration")

        else:

            if identifier in self.targets:

                for key in self.targets[identifier].regulatory_interactions.keys():
                    self.remove_regulatory_interaction(key)

                if identifier in self.targets:
                    del self.targets[identifier]

            if identifier in self.regulators:

                for key in self.regulators[identifier].regulatory_interactions.keys():
                    self.remove_regulatory_interaction(key)

                if identifier in self.regulators:
                    del self.regulators[identifier]

    def update_aliases_from_tabular_format_file(self, aliases_file,
                                                sep=None,
                                                id_col=None,
                                                aliases_cols=None,
                                                header=None):

        if not isinstance(aliases_file, str):
            raise TypeError("File or file path must be a string")

        df = read_tabular_aliases(aliases_file, sep=sep, id_col=id_col, aliases_cols=aliases_cols, header=header)

        aliases_cols = [col for col in df.columns if col != 'ids']

        for row in df.index:

            new_aliases = list(df.loc[row, aliases_cols])

            if row in self.regulatory_variables:
                self.regulatory_variables[row].aliases = new_aliases

    # TODO: Context manager for storing history. Useful for temporary changes such as KO

    # TODO: the dictionaries used in the regulatory model and variables must have some of the methods re-written,
    #  namely the set attribute


def solution_decode(solution):

    decode = {
        True: 1,
        False: 0,
        1: 1,
        0: 0,
        Zero: 0,
        One: 1,
        BooleanFalse: 0,
        BooleanTrue: 1,
        None: None
    }

    if solution in decode:
        return decode[solution]

    else:
        return solution



from mewpy.problems.problem import AbstractOUProblem
from mewpy.utils.parsing import Boolean, GeneEvaluator, build_tree
from mewpy.utils.constants import EAConstants
from mewpy.regulation.variable import RegulatoryVariable
from collections import OrderedDict
import numpy as np
import pandas as pd
import math


def load_optram(gene_filename, tf_filename, matrix_filename, gene_prefix=''):
    """
    Loads a OptRAM regulatory model from csv files:

        gene_file columns: "Name;id"
        tf_file columns: "Name;Expression"
        matrix_file: genes as rows and tfs as columns
    """
    df_genes = pd.read_csv(gene_filename)
    df_TFs = pd.read_csv(tf_filename)
    mat = pd.read_csv(matrix_filename, header=None)

    genes = OrderedDict()
    for index, row in df_genes.iterrows():
        genes[gene_prefix+row['Name']
              ] = RegGene(gene_prefix+row['Name'], index, row['id'])
    tfs = OrderedDict()
    for index, row in df_TFs.iterrows():
        tf = TF(row['Name'], index, row['Expression'])
        tfs[row['Name']] = tf

    model = OptRAMRegModel(genes, tfs, mat)
    return model


class RegGene(RegulatoryVariable):
    """Genes included in the regulatory model

       args:
        name (str): the gene identifier
        row (int): the associated row in the regulatory model
        id (int): OptRAM ID
        cbm_name (str): the gene corresponding name in the constraint base model (G_XXXXX)

    """

    def __init__(self, name, row, optramid, aliases=None):
        super().__init__(name=name, aliases=aliases)
        self.row = row
        self.optramid = optramid


class TF(RegulatoryVariable):
    """Transcription factor

    args:
        name (str): the TF identifier
        column (int): the associated column in the regulatory model
        expression (str): the TF expression (By default 1)
    """

    def __init__(self, name, column, expression=1):
        super().__init__(name=name)
        self.column = column
        self.expression = expression


class OptRAMRegModel:
    def __init__(self, genes, tfs, regnet):
        """
        OptRAM regulatory network model.

        :param genes: (dic) A dictionary of Gene objects.
        :param tfs: (dic) A dictionary of transcription factors (TF) objects.
        :param regnet: (DataFrame) A panda dataframe containing a matrix of Genes TFs coefficients.
        """
        self.genes = genes
        self.tfs = tfs
        # panda (genes x TFs)
        # TODO: maybe use a list of lists instead of a panda dataframe
        self.regnet = regnet
        self.tf_expression = [0]*len(self.tfs)
        for _, tf in self.tfs.items():
            self.tf_expression[tf.column] = tf.expression


class OptRamProblem(AbstractOUProblem):

    def __init__(self, model, fevaluation, regmodel, **kwargs):
        """
        EA OptRam problem
        """
        super(OptRamProblem, self).__init__(model, fevaluation, **kwargs)
        self.regmodel = regmodel
        # ignore user defined target list
        self._trg_list = None
        # GPR operators
        self._operators = None
        # Reset default OU levels to OptRAM levels if none are provided
        self.levels = kwargs.get('levels', EAConstants.OPTRAM_LEVELS)

    def _build_target_list(self):
        """ The EA target list is the combination [mGene]+[TFs]
        """
        self._trg_list = []
        self._trg_list.extend(list(self.regmodel.genes.keys()))
        self._trg_list.extend(list(self.regmodel.tfs.keys()))

    def target_exp(self, candidate):

        mgenes_p = {}
        # TFs expression vector
        tf_exp_v = self.regmodel.tf_expression.copy()
        # keeps track of the TF whose expression is altered by the candidate
        tf_altered = [False] * len(tf_exp_v)
        for idx, lv_idx in candidate:
            try:
                target = self.target_list[idx]
                lv = self.levels[lv_idx]

                if idx >= len(self.regmodel.genes):
                    # TF expression level encoded in the candidate
                    tf = self.regmodel.tfs[target]
                    c = tf.column
                    tf_exp_v[c] *= lv
                    tf_altered[c] = True
                else:
                    # gene expression level encoded in the candidate
                    #
                    mgenes_p[target] = lv
            except IndexError:
                raise IndexError("Index out of range")

        # update gene expression level with TFs
        # gene expression is p=2^sum(coeff*log2 tfexpr)

        log_tfexpr = [math.log2(x) for x in tf_exp_v]

        for g in self.regmodel.genes:
            g_idx = self.regmodel.genes[g].row
            coeff = abs(self.regmodel.regnet.iloc[g_idx].to_numpy())
            # only genes with altered TFs have their expression level updated
            if np.dot(coeff != 0, tf_altered):
                p = 2 ** np.dot(coeff, log_tfexpr)
                mgenes_p[g] = p
        return mgenes_p

    def decode(self, candidate):
        gr_constraints = OrderedDict()
        mgenes_p = self.target_exp(candidate)
        # Evaluate gpr.
        if not self._operators:
            self._operators = (lambda x, y: min(x, y), lambda x, y: max(x, y))
        evaluator = GeneEvaluator(
            mgenes_p, self._operators[0], self._operators[1])
        for rxn_id in self.simulator.reactions:
            if self.simulator.get_gpr(rxn_id):
                gpr = str(self.simulator.get_gpr(rxn_id))
                tree = build_tree(gpr, Boolean)
                # Apply the operators to obtain a level for the reaction.
                # If a gene in the gpr has no associated level,
                # its factor is 1 (see mewpy.util.parsing.GeneEvaluator)
                lv = tree.evaluate(evaluator.f_operand, evaluator.f_operator)
                # Adds the reaction constraint.
                rev_rxn = self.simul_context.reverse_reaction(rxn_id)
                # Skips if the reverse reaction was already processed.
                if rev_rxn and rev_rxn in gr_constraints.keys():
                    continue
                elif lv < 0:
                    raise ValueError("All UO levels should be positive")
                elif lv == 1:
                    # No constraints are added.
                    continue
                else:
                    gr_constraints.update(
                        self.reaction_constraints(rxn_id, lv))
        return gr_constraints
